# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Classes representing file information.
"""

from collections import defaultdict
from dataclasses import dataclass
import logging
import os
from datetime import datetime, timezone
import re
import configparser
from typing import Callable, Union

from atbu.mp_pipeline.mp_global import get_verbosity_level
from atbu.common.exception import (
    exc_to_string,
    AtbuException,
    FileChangedWhileCalculatingHash,
    InvalidStateError,
    StateChangeDisallowedError,
    AlreadyUsedError,
)
from atbu.common.hasher import (
    Hasher,
    HasherDefinitions,
)

from ..backup.exception import (
    PersistentFileInfoError,
    InvalidPersistentFileInfoError,
    PersistentFileInfoVersionMismatch,
)
from ..backup.constants import *
from ..backup.global_hasher import GlobalHasherDefinitions

CHANGE_DETECTION_TYPE_DATESIZE = "datesize"
CHANGE_DETECTION_TYPE_DIGEST = "digest"
CHANGE_DETECTION_TYPE_FORCE = "force"
CHANGE_DETECTION_CHOICES = [
    CHANGE_DETECTION_TYPE_DATESIZE,
    CHANGE_DETECTION_TYPE_DIGEST,
    CHANGE_DETECTION_TYPE_FORCE,
]


def has_atbu_extension(path):
    ext: str = os.path.splitext(path)[1]
    return ext.lower() in ATBU_SKIP_EXTENSIONS


class FileInformation:
    """Each instance represents information about a file."""

    FILE_READ_SIZE_5MB = 1024 * 1024 * 5
    path_stamp_failure_warning_tracker = set()

    def __init__(self, path: str):
        self._implicit_refresh_allowed = True
        self.path = path
        self.dirname = os.path.dirname(path)
        self.basename = os.path.basename(path)
        self.basename_no_ext, self.ext = os.path.splitext(self.basename)
        self._size_in_bytes = None
        self._modified_time_posix = None
        self._accessed_time_posix = None  # Retained to allow use of utime.
        self.digests = None

    def refresh(self):
        self.refresh_stat_info()
        self.refresh_digests(GlobalHasherDefinitions().create_hasher())

    @property
    def nc_path(self):
        return os.path.normcase(self.path)

    @property
    def size_in_bytes(self):
        if self._size_in_bytes is None:
            self.refresh_stat_info()
        return self._size_in_bytes

    @size_in_bytes.setter
    def size_in_bytes(self, value):
        self._size_in_bytes = value

    @property
    def modified_time_posix(self):
        if self._modified_time_posix is None:
            self.refresh_stat_info()
        return self._modified_time_posix

    @modified_time_posix.setter
    def modified_time_posix(self, v):
        self._modified_time_posix = v

    @property
    def accessed_time_posix(self):
        if self._accessed_time_posix is None:
            self.refresh_stat_info()
        return self._accessed_time_posix

    @accessed_time_posix.setter
    def accessed_time_posix(self, v):
        self._accessed_time_posix = v

    @property
    def implicit_refresh_allowed(self):
        return self._implicit_refresh_allowed

    @implicit_refresh_allowed.setter
    def implicit_refresh_allowed(self, value):
        self._implicit_refresh_allowed = value

    def calculate_hashes(self, hasher: Hasher = None, max_attempts=None) -> Hasher:
        if not self._implicit_refresh_allowed:
            raise StateChangeDisallowedError(
                f"Value should be set from persisted state, refresh disallowed."
            )
        if max_attempts is not None and max_attempts <= 0:
            raise ValueError(
                f"max_attempts must be None for forever, "
                f"or a positive number indicating max attempts."
            )
        if hasher is None:
            hasher = GlobalHasherDefinitions().create_hasher()
        while True:
            modified_time_before, size_in_bytes_before = self.refresh_stat_info()
            with open(self.path, "rb") as file:
                while True:
                    data = file.read(FileInformation.FILE_READ_SIZE_5MB)
                    if not data:
                        break
                    hasher.update_all(data)
            modified_time_after, size_in_bytes_after = self.refresh_stat_info()
            if (
                modified_time_before == modified_time_after
                and size_in_bytes_before == size_in_bytes_after
            ):
                # File did not change (from OS perspective).
                break
            if max_attempts is None:
                # try forever
                continue
            max_attempts -= 1
            if max_attempts <= 0:
                raise FileChangedWhileCalculatingHash(
                    f"The file changed while calculating hashes: {self.path}"
                )
        return hasher

    def refresh_stat_info(self) -> tuple[float, int]:
        if not self._implicit_refresh_allowed:
            raise StateChangeDisallowedError(
                f"Value should be set from persisted state, refresh disallowed."
            )
        sr: os.stat_result = os.stat(self.path)
        self._size_in_bytes = sr.st_size
        self._modified_time_posix = sr.st_mtime
        self._accessed_time_posix = sr.st_atime
        return self._modified_time_posix, self._size_in_bytes

    def refresh_digests(self, hasher: Hasher = None, max_attempts: int = None) -> dict:
        hasher = self.calculate_hashes(hasher=hasher, max_attempts=max_attempts)
        self.digests = hasher.get_hexdigests()
        return self.digests

    @property
    def has_primary_digest(self):
        """True if digest already generated, False if not."""
        return (
            self.digests is not None
            and self.digests.get(
                GlobalHasherDefinitions().get_primary_hashing_algo_name()
            )
            is not None
        )

    @property
    def primary_digest(self) -> str:
        if self.digests is None:
            self.refresh_digests()
        return self.digests[GlobalHasherDefinitions().get_primary_hashing_algo_name()]

    @primary_digest.setter
    def primary_digest(self, value):
        if self.digests is None:
            self.digests = {}
        self.digests[GlobalHasherDefinitions().get_primary_hashing_algo_name()] = value

    @property
    def primary_digest_algo_name(self) -> str:
        return GlobalHasherDefinitions().get_primary_hashing_algo_name()

    def _get_date_stamp_ISO8601(self, posix_timestamp: float, tz=None):
        if posix_timestamp is None:
            self.refresh_stat_info()
        return FileInformation.get_datetime_stamp_ISO8601(
            datetime.fromtimestamp(posix_timestamp, tz=tz)
        )

    @property
    def modified_date_stamp_ISO8601_local(self):
        return self._get_date_stamp_ISO8601(posix_timestamp=self.modified_time_posix)

    @property
    def modified_date_stamp_ISO8601_utc(self):
        """Obtain a textual user-friendly ISO8601 UTC stamp.

        WARNING: Caller's should only use this for debugging or easy viewing of
        time stamps in cases where the stamps are likely to be available.
        Python can raise OSError 22 Invalid Argument if a POSIX stamp is
        out of a certain range. For example, 130964362929 on Windows, which
        is within the year 6000+.

        Returns:
            str: An ISO8601 UTC stamp or an error message if a failure occurs.
        """
        try:
            return self._get_date_stamp_ISO8601(
                posix_timestamp=self.modified_time_posix, tz=timezone.utc
            )
        except OSError as ex:
            if self.path not in FileInformation.path_stamp_failure_warning_tracker:
                FileInformation.path_stamp_failure_warning_tracker.add(self.path)
                logging.warning(
                    f"modified_date_stamp_ISO8601_utc: "
                    f"Failed to get ISO8601 stamp. "
                    f"You can likely ignore this warning. "
                    f"path={self.path}: {exc_to_string(ex)}"
                )
            return (
                f"failed-for-stamp:"
                f"errno={ex.winerror},"
                f"msg={ex.strerror},"
                f"mt_posix={self.modified_time_posix}"
            )

    @property
    def accessed_date_stamp_ISO8601_local(self):
        return self._get_date_stamp_ISO8601(posix_timestamp=self.accessed_time_posix)

    @property
    def accessed_date_stamp_ISO8601_utc(self):
        return self._get_date_stamp_ISO8601(
            posix_timestamp=self.accessed_time_posix, tz=timezone.utc
        )

    @staticmethod
    def get_datetime_stamp_v01(dt):
        return dt.strftime("%Y/%m/%d-%H:%M:%S")

    @staticmethod
    def get_datetime_stamp_ISO8601(dt: datetime):
        return dt.isoformat(timespec="microseconds")

    @staticmethod
    def get_datetime_from_ISO8601(iso8601: str):
        return datetime.fromisoformat(iso8601)

    def to_serialization_dict(self) -> dict:
        d = {
            "_type": "FileInformation",
            "path": self.path,
            # The original POSIX timestamp returned by Python's stat.
            "lastmodified": self._modified_time_posix,
            "lastaccessed": self._accessed_time_posix,
            # The stamp rounds to microseconds (given datetime instances have
            # microsecond accuracy to whatever degree supported by the platform).
            "lastmodified_stamp": self.modified_date_stamp_ISO8601_utc,
            "size_in_bytes": self.size_in_bytes,
            "digests": self.digests,
        }
        return d

    def from_serialization_dict(self, d: dict):
        # By default, disallowed refresh of loaded persisted state.
        self._implicit_refresh_allowed = False
        self.path = d["path"]
        # Follwoing saves original POSIX stamp for potential usage in
        # debugging and/or later time-handling code mods.
        self._modified_time_posix = d["lastmodified"]
        self._accessed_time_posix = d["lastaccessed"]
        # This sets self._modified_time to "datetime microsecond"
        # truncated original.
        self.size_in_bytes = d["size_in_bytes"]
        self.digests = d["digests"]


class FileInformationPersistent(FileInformation):
    """Like FileInformation, instances of FileInformationPersistent
    represent information about a file, but are instances that represent
    file information as obtained from one of several mechanisms...

        Originally, the information is likely obtained from a file
        within a file system.

        The original information can be stored in per-file .atbu file
        information files, where it can be later retrieved for use in
        comparing with other files or even with its own current state.

        The original information can also be stored in a file information
        database (json db) where it can be compared with either another
        json db or a live file system.

    In all the above cases, once information is retrieved, whether from
    the file system, a json db, or a .atbu file information file, the
    information is stored in instances of FileInformationPersistent in
    a "firewalled" manner from the live OS information.

    Therefore, any testing of FileInformationPersistent should ensure that
    file information read from any other source is never polluted from live
    OS information until such time as someone calls a method on the instance
    (i.e., refresh_info_from_phys_file) to do so.

    With all that said, keep in mind quite simply that FileInformation and
    FileInformationPersistent are different. FileInformation deals with
    information just read from the file system (not info saved in a db or
    .atbu file), whereas FileInformationPersistent, aside from its initial
    sampling of information from the file system, never deals with live OS
    data until commanded to do so.
    """

    def __init__(
        self,
        path,
        info_current=None,
        info_history=None,
        is_loaded_from_db: bool = False,
    ):
        self.version = None  # added for persistence v2.
        self.is_loaded_from_db = is_loaded_from_db  # Not persisted.
        if info_current is None:
            info_current = {}
        if info_history is None:
            info_history = []
        super().__init__(path=path)
        self.config_basename = self.basename + ATBU_PERSISTENT_INFO_EXTENSION
        self.info_data_file_path = os.path.join(self.dirname, self.config_basename)
        self.info_current = info_current
        digest_from_current = info_current.get(
            GlobalHasherDefinitions().get_primary_hashing_algo_name()
        )
        if digest_from_current:
            self.primary_digest = digest_from_current
        self.info_history = info_history
        # Not yet persisted but updated via update_stat_info
        # for comparison or usage generally before such updates.
        self.cached_digests = None
        self.cached_statinfo = None

    @property
    def size_in_bytes_cached(self) -> int:
        """Return the current size in bytes."""
        if not self.has_cached_size_in_bytes:
            raise KeyError("'sizeinbytes' not cached yet, refresh required first.")
        return int(self.cached_statinfo["sizeinbytes"])

    @property
    def has_cached_size_in_bytes(self) -> bool:
        return (
            self.cached_statinfo is not None and "sizeinbytes" in self.cached_statinfo
        )

    @property
    def size_in_bytes(self):
        """Return the current size in bytes."""
        if "sizeinbytes" not in self.info_current:
            raise KeyError("SizeInBytes not saved or loaded to/from file information.")
        return int(self.info_current["sizeinbytes"])

    @property
    def modified_date_stamp_local_cached(self) -> str:
        if "lastmodified" not in self.cached_statinfo:
            raise KeyError("lastmodified not cached yet, refresh required first.")
        return self._get_date_stamp_ISO8601(
            posix_timestamp=float(self.cached_statinfo["lastmodified"]),
        )

    @property
    def has_cached_modified_date(self) -> bool:
        return (
            self.cached_statinfo is not None and "lastmodified" in self.cached_statinfo
        )

    @property
    def modified_date_stamp_local(self) -> str:
        if "lastmodified" not in self.info_current:
            raise KeyError("lastmodified not cached yet, refresh required first.")
        return self._get_date_stamp_ISO8601(
            posix_timestamp=float(self.info_current["lastmodified"]),
        )

    def get_digest_cached(self, hashing_algo_name=None) -> str:
        if hashing_algo_name is None:
            hashing_algo_name = (
                GlobalHasherDefinitions().get_primary_hashing_algo_name()
            )
        return self.cached_digests.get(hashing_algo_name.lower())

    def has_cached_digest(self, hashing_algo_name=None) -> bool:
        if hashing_algo_name is None:
            hashing_algo_name = (
                GlobalHasherDefinitions().get_primary_hashing_algo_name()
            )
        return (
            self.cached_digests is not None and hashing_algo_name in self.cached_digests
        )

    def get_current_digest(self, hashing_algo_name=None) -> str:
        if hashing_algo_name is None:
            hashing_algo_name = (
                GlobalHasherDefinitions().get_primary_hashing_algo_name()
            )
        return self.info_current.get(hashing_algo_name.lower())

    def is_modified_date_or_size_changed(self):
        if len(self.info_current) == 0:
            return True
        if "lastmodified" not in self.info_current:
            return True
        if "sizeinbytes" not in self.info_current:
            return True
        self.refresh_stat_info()
        if float(self.info_current["lastmodified"]) != float(
            self.cached_statinfo["lastmodified"]
        ):
            return True
        if int(self.info_current["sizeinbytes"]) != int(
            self.cached_statinfo["sizeinbytes"]
        ):
            return True
        return False

    def is_primary_digest_changed(self, hasher: Hasher) -> bool:
        if len(self.info_current) == 0:
            return True
        primary_hasher_name = hasher.get_primary_hashing_algo_name()
        if primary_hasher_name not in self.info_current:
            return True
        self.refresh_digests(hasher)
        if (
            self.info_current[primary_hasher_name]
            != self.cached_digests[primary_hasher_name]
        ):
            return True
        return False

    def refresh_info_from_phys_file(self, hashers):
        if self.cached_digests is None:
            self.refresh_digests(hashers)
        if self.cached_statinfo is None:
            self.refresh_stat_info()
        info_new = self.cached_statinfo | self.cached_digests
        if self.info_current == info_new:
            return
        self.info_current = info_new
        self.info_history.append(info_new)

    def clear_cached_info(self):
        self.cached_digests = None
        self.cached_statinfo = None

    def info_data_file_exists(self):
        return os.path.exists(self.info_data_file_path)

    def _create_config_parser(self):
        return configparser.ConfigParser(interpolation=None)

    def read_info_data_file(self):
        if not self.info_data_file_exists():
            raise FileNotFoundError(
                f"Config does not exist: {self.info_data_file_path}"
            )
        self.info_current = {}
        self.info_history = []
        cp = self._create_config_parser()
        cp.read(self.info_data_file_path)
        if "MAIN" not in cp.sections():
            raise InvalidPersistentFileInfoError(
                f"The persistent info does not have "
                f"the expected MAIN section: "
                f"{self.info_data_file_path}"
            )
        if "INFO.CURRENT" in cp.sections():
            self.info_current = dict(cp["INFO.CURRENT"])
        re_info_history = re.compile(r"^INFO.\d{4}$")
        history_sections = list(sorted(filter(re_info_history.match, cp.sections())))
        for history_section in history_sections:
            self.info_history.append(dict(cp[history_section]))
        self.version = dict(cp["MAIN"])["version"]
        if self.version != "2":
            raise PersistentFileInfoVersionMismatch(
                f"The persistent info file version is mismatched: "
                f"v={self.version} {self.info_data_file_path}"
            )
        self.primary_digest = self.get_current_digest()

    def write_info_data_file(self):
        cp = self._create_config_parser()
        cp["MAIN"] = {
            "Version": ATBU_PERSISTENT_INFO_VERSION_STRING,
            "BaseName": self.basename,
        }
        cp["INFO.CURRENT"] = self.info_current
        for index, info_item in enumerate(self.info_history):
            cp[f"INFO.{index:04d}"] = info_item
        with open(self.info_data_file_path, "w", encoding="utf-8") as cf:
            cp.write(cf)

    def update_info_data_file_to_latest_version(self, hashers):
        # There has been only version 1 and 2, current being 2.
        # Following logic is therefore to upgrade from 1 to 2.
        # Modify accordingly for subsequent upgrades.

        if self.version != "0.1":
            # Caller should have attempted read_info_data_file
            # which sets version.
            raise PersistentFileInfoError(
                f"update_info_data_file_to_latest_version cannot "
                f"be called for anything but version 1.0."
            )

        self.clear_cached_info()
        self.refresh_stat_info()

        stamp = FileInformation.get_datetime_stamp_v01(
            datetime.fromtimestamp(self._modified_time_posix, tz=timezone.utc)
        )

        is_changed = (
            len(self.info_current) == 0
            or "lastmodified" not in self.info_current
            or "sizeinbytes" not in self.info_current
            or self.info_current["lastmodified"] != stamp
            or int(self.info_current["sizeinbytes"])
            != int(self.cached_statinfo["sizeinbytes"])
        )

        if is_changed:
            # Perform full update including digests if either of
            # date/size has changed.
            self.refresh_info_from_phys_file(hashers)
            self.write_info_data_file()
            self.read_info_data_file()
            return

        # Otherwise, perform a quicker upgrade of simply stat info.
        # Merge in updated state info while retaining already calculated digests.
        self.info_current = self.info_current | self.cached_statinfo
        self.info_history.append(self.info_current)
        self.write_info_data_file()
        self.read_info_data_file()

    def refresh_stat_info(self) -> tuple[float, int]:
        mod_time, size = super().refresh_stat_info()
        self.cached_statinfo = {
            "sizeinbytes": self._size_in_bytes,
            "lastmodified": self.modified_time_posix,
            "lastmodified_stamp": self.modified_date_stamp_ISO8601_utc,
        }
        return mod_time, size

    def refresh_digests(self, hasher: Hasher = None, max_attempts: int = None):
        # For the purposes of this class, if the cached digests
        # are updated, the cached dict date/size info should be
        # matching that which existed which hashes sampled.
        self.refresh_stat_info()
        r = super().refresh_digests(hasher=hasher, max_attempts=max_attempts)
        # Digests are already in dict and therefore config
        # file format, just copy them to our digest info cache.
        # This is unlike _update_cached_stat_info converts to dict.
        self.cached_digests = self.digests
        return r

    def __str__(self) -> str:
        s = []
        s.append(f"path={self.path}")
        s.append(f"config_path={self.info_data_file_path}")
        s.append("info_current...")
        if len(self.info_current) > 0:
            for k, v in self.info_current.items():
                s.append(f"  {k}={v}")
        else:
            s.append("  Empty")
        s.append("info_history...")
        if len(self.info_history) > 0:
            for index, item in enumerate(self.info_history):
                s.append(f"  INFO.{index:04d}:")
                for k, v in item.items():
                    s.append(f"    {k}={v}")
        else:
            s.append("  Empty")
        return "\n".join(s)

    def to_serialization_dict(self) -> dict:
        return {
            "_type": "FileInformationPersistent",
            "path": self.path,
            "info_current": self.info_current,
            "info_history": self.info_history,
        }


@dataclass
class SneakyCorruptionPotential:
    file_info: FileInformationPersistent
    old_size_in_bytes: int
    cur_size_in_bytes: int
    old_modified_time: str
    cur_modified_time: str
    old_digest: str
    cur_digest: str


class LocationFileInfoUpdater:
    def __init__(
        self,
        file_info_list: list[FileInformationPersistent],
        update_stale: bool,
        change_detection_type: str,
        per_file_persistence: bool = True,
        whatif: bool = False,
        hasher_defs: HasherDefinitions = None,
    ):
        if not isinstance(file_info_list, list):
            raise ValueError(
                "file_info_list must be an Interable of FileInformationPersistent instances."
            )
        if hasher_defs is None:
            hasher_defs = GlobalHasherDefinitions()
        if not change_detection_type:
            raise ValueError(
                "change_detection_type must be one of datesize, digest, force."
            )
        self.whatif = whatif
        self.whatif_str = "(--whatif) " if self.whatif else ""
        self.file_info_list = file_info_list
        self.hasher_defs = hasher_defs
        self.update_stale = update_stale
        self.change_detection_type = change_detection_type
        self.per_file_persistence = per_file_persistence
        self.primary_hasher_name = hasher_defs.get_primary_hashing_algo_name()
        self.skipped_files: list[tuple[FileInformationPersistent, str]] = []
        # Counters
        self.total_files = 0
        self.total_file_info_created = 0
        self.total_file_info_updated = 0
        self.total_files_stale_info_skipped = 0
        self.total_files_no_update_required = 0
        self.total_files_without_primary_digest = 0
        self.sneaky_corruption_potentials: list[SneakyCorruptionPotential] = []
        self.is_used = False

    def _process_file_info(self, file_info: FileInformationPersistent):
        self.total_files += 1
        some_updating_occurred = False
        info_data_file_existed = False
        try:
            # If a file info came from a DB file, it effectively
            # represents a file info read from disk.
            if not file_info.is_loaded_from_db:
                if self.per_file_persistence:
                    file_info.read_info_data_file()
                else:
                    # This is a "new file" for a DB-only scan because
                    # is_loaded_from_db==False and per_file_persistence==False.
                    # Note, combo scans, scans for both per-dir and per-file for
                    # a location, will perform a per-file scan whose results
                    # feed the DB, but not so in this case.
                    raise FileNotFoundError(
                        f"New DB item FileNotFoundError: {file_info.path}"
                    )
            info_data_file_existed = True
        except FileNotFoundError:
            if not self.update_stale:
                logging.warning(f"File info does not exist, skipping {file_info.path}.")
                self.total_files_stale_info_skipped += 1
                self.skipped_files.append(
                    (file_info, f"The {ATBU_PERSISTENT_INFO_EXTENSION} info is missing")
                )
                return
            logging.info(f"{self.whatif_str}Creating info for {file_info.path}...")
            if not self.whatif:
                file_info.refresh_info_from_phys_file(self.hasher_defs.create_hasher())
                if self.per_file_persistence:
                    file_info.write_info_data_file()
                self.total_file_info_created += 1
            some_updating_occurred = True
        except PersistentFileInfoVersionMismatch as ex:
            if not self.update_stale:
                raise AtbuException(
                    f"Cannot upgrade persistent file info because "
                    f"option to update not specified. {exc_to_string(ex)}"
                ).with_traceback(ex.__traceback__) from ex
            logging.info(f"{self.whatif_str}Creating info for {file_info.path}...")
            if not self.whatif:
                file_info.update_info_data_file_to_latest_version(
                    self.hasher_defs.create_hasher()
                )
                self.total_file_info_updated += 1
            some_updating_occurred = True
        primary_digest = file_info.get_current_digest(self.primary_hasher_name)
        if primary_digest is None:
            self.total_files_without_primary_digest += 1
            self.skipped_files.append(
                (file_info, f"No '{self.primary_hasher_name}' primary digest found")
            )
            logging.warning(
                f"No '{self.primary_hasher_name}' primary digest found, skipping {file_info.path}"
            )
            return
        is_changed = False
        if get_verbosity_level() > 0:
            logging.info(f"Checking for changes to {file_info.path}...")
        if self.change_detection_type == CHANGE_DETECTION_TYPE_DATESIZE:
            is_changed = file_info.is_modified_date_or_size_changed()
        elif self.change_detection_type == CHANGE_DETECTION_TYPE_DIGEST:
            is_changed = file_info.is_primary_digest_changed(
                self.hasher_defs.create_hasher()
            )
        elif self.change_detection_type == CHANGE_DETECTION_TYPE_FORCE:
            is_changed = True
        else:
            raise InvalidStateError(
                f"Unexpected change_detection_type={self.change_detection_type}."
            )
        if is_changed:
            logging.info(f"Change detected: {file_info.path}:")
        if is_changed or get_verbosity_level() > 1:
            if file_info.has_cached_size_in_bytes:
                logging.info(f"    cur size={file_info.size_in_bytes_cached}")
                logging.info(f"    old size={file_info.size_in_bytes}")
            elif self.change_detection_type != CHANGE_DETECTION_TYPE_FORCE:
                logging.info(f"    (size not checked)")
            if file_info.has_cached_modified_date:
                logging.info(
                    f"    cur time={file_info.modified_date_stamp_local_cached}"
                )
                logging.info(f"    old time={file_info.modified_date_stamp_local}")
            elif self.change_detection_type != CHANGE_DETECTION_TYPE_FORCE:
                logging.info(f"    (time not checked)")
            if file_info.has_cached_digest():
                logging.info(f"    cur digest={file_info.get_digest_cached()}")
                logging.info(f"    old digest={file_info.get_current_digest()}")
            elif self.change_detection_type != CHANGE_DETECTION_TYPE_FORCE:
                logging.info(f"    (digest not checked)")
            if (
                file_info.has_cached_size_in_bytes
                and file_info.has_cached_modified_date
                and file_info.has_cached_digest()
            ):
                if (
                    file_info.size_in_bytes_cached == file_info.size_in_bytes
                    and file_info.modified_date_stamp_local_cached
                    == file_info.modified_date_stamp_local
                    and file_info.get_digest_cached() != file_info.get_current_digest()
                ):
                    logging.warning(
                        f"WARNING: Potential bitrot or other sneaky corruption: {file_info.path}"
                    )
                    self.sneaky_corruption_potentials.append(
                        SneakyCorruptionPotential(
                            file_info=file_info,
                            old_size_in_bytes=file_info.size_in_bytes,
                            cur_size_in_bytes=file_info.size_in_bytes_cached,
                            old_modified_time=file_info.modified_date_stamp_local,
                            cur_modified_time=file_info.modified_date_stamp_local_cached,
                            old_digest=file_info.get_current_digest(),
                            cur_digest=file_info.get_digest_cached(),
                        )
                    )
        if is_changed:
            if not self.update_stale:
                self.total_files_stale_info_skipped += 1
                self.skipped_files.append(
                    (file_info, f"The {ATBU_PERSISTENT_INFO_EXTENSION} info is stale")
                )
                logging.warning(
                    f"File has changed, file info stale, skipping {file_info.path}."
                )
                return
            logging.info(f"{self.whatif_str}Updating file info for {file_info.path}...")
            if not self.whatif:
                file_info.refresh_info_from_phys_file(self.hasher_defs.create_hasher())
                primary_digest = file_info.get_current_digest(self.primary_hasher_name)
                if self.per_file_persistence:
                    file_info.write_info_data_file()
            some_updating_occurred = True
            if file_info.is_modified_date_or_size_changed() and not self.whatif:
                self.skipped_files.append(
                    (
                        file_info,
                        f"The {ATBU_PERSISTENT_INFO_EXTENSION} file info "
                        f"remains stale after update attempt",
                    )
                )
                self.total_files_stale_info_skipped += 1
                logging.warning(
                    f"The {ATBU_PERSISTENT_INFO_EXTENSION} file info remains "
                    f"stale after update attempt, skipping {file_info.path}."
                )
                return
            self.total_file_info_updated += 1
        if not some_updating_occurred:
            self.total_files_no_update_required += 1
        if some_updating_occurred:
            added_changed_str = "updated" if info_data_file_existed else "added"
            if self.per_file_persistence:
                logging.info(
                    f"The {ATBU_PERSISTENT_INFO_EXTENSION} file info was {added_changed_str}: "
                    f"path={file_info.path} "
                    f"{file_info.primary_digest_algo_name}={file_info.get_current_digest()}"
                )
                logging.debug(
                    f"The {ATBU_PERSISTENT_INFO_EXTENSION} file info was "
                    f"{added_changed_str}: {file_info}"
                )
            else:
                logging.info(
                    f"The file info was {added_changed_str}: "
                    f"path={file_info.path} "
                    f"{file_info.primary_digest_algo_name}={file_info.get_current_digest()}"
                )
                logging.debug(f"The file info was {added_changed_str}: {file_info}")
        elif get_verbosity_level() > 0:
            if self.per_file_persistence:
                logging.info(
                    f"The {ATBU_PERSISTENT_INFO_EXTENSION} file info was up to date: "
                    f"path={file_info.path} "
                    f"{file_info.primary_digest_algo_name}={file_info.get_current_digest()}"
                )
                logging.debug(
                    f"The {ATBU_PERSISTENT_INFO_EXTENSION} file info was up to date: {file_info}"
                )
            else:
                logging.info(
                    f"The file info was up to date: "
                    f"path={file_info.path} "
                    f"{file_info.primary_digest_algo_name}={file_info.get_current_digest()}"
                )
                logging.debug(f"The file info was up to date: {file_info}")

        # logging.info(f"Have file w/good info for {file_info.path}")
        return

    def update(self):
        if self.is_used:
            raise AlreadyUsedError(f"This instance already used to update.")
        self.is_used = True
        for file_info in self.file_info_list:
            self._process_file_info(file_info)


def get_loc_files_as_ncloc_to_object(
    locations: Union[str, list[str]],
    include_pattern: re.Pattern = None,
    exclude_atbu: bool = True,
    factory: Callable[[str, str], object] = lambda loc_root, file_path: str(file_path),
) -> dict[str, list[str]]:
    """Get each location's files paired with the root location leading to each discovered file.
    This returns a dictionary of location to files for each such location. For each file,
    the factory is called as follows...

        factory(location_root_path, path_to_file)

    ...where the factory should return an instance of the desired class.

    The default factory creates a string of the path to the file.
    """
    if isinstance(locations, str):
        locations = [str(locations)]
    results: dict = defaultdict(list)
    for location in locations:
        for root, dirs, files in os.walk(location):  # pylint: disable=unused-variable
            for file in files:
                # Skip .atbu extension files. We get to the info via non-atbu files.
                if exclude_atbu and has_atbu_extension(file):
                    continue
                if include_pattern and not include_pattern.match(file):
                    continue
                o = factory(location, os.path.join(root, file))
                if o:
                    results[os.path.normcase(location)].append(o)
    return results


def get_loc_fileobj_list(
    locations: Union[str, list[str]],
    include_pattern: re.Pattern = None,
    exclude_atbu: bool = True,
    factory: Callable[[str, str], object] = lambda loc_root, file_path: str(file_path),
) -> list[str]:
    """Get all files as an object, which is by default the file's path."""
    dict_ncloc_to_file_list = get_loc_files_as_ncloc_to_object(
        locations=locations,
        include_pattern=include_pattern,
        exclude_atbu=exclude_atbu,
        factory=factory,
    )
    file_list = [v for l in dict_ncloc_to_file_list.values() for v in l]
    return file_list


def get_loc_persistent_file_info(
    locations: Union[str, list[str]],
    include_pattern: re.Pattern = None,
) -> list[FileInformationPersistent]:
    file_info_persistent_list: list[FileInformationPersistent] = get_loc_fileobj_list(
        locations=locations,
        include_pattern=include_pattern,
        exclude_atbu=True,
        factory=lambda loc_root, file_path: FileInformationPersistent(path=file_path),
    )
    return file_info_persistent_list


def is_file_info_list_bad_state(
    primary_hasher_name, file_info_list: list[FileInformationPersistent]
):
    primary_digest = file_info_list[0].get_current_digest(primary_hasher_name)
    size_in_bytes = file_info_list[0].size_in_bytes
    state_error = False
    for file_info in file_info_list[1:]:
        other_primary_digest = file_info.get_current_digest(primary_hasher_name)
        if primary_digest != other_primary_digest:
            logging.error(
                f"Unexpected list state error, digest mismatch "
                f"'{primary_digest}' != '{other_primary_digest}'"
            )
            for file_info2 in file_info_list:
                logging.error(f"item related to bad state: {file_info2}")
            state_error = True
            break
        if size_in_bytes != file_info.size_in_bytes:
            logging.error(
                "Unexpected size mismatch encountered, "
                f"potential collision, skipping the following file:"
            )
            for file_info2 in file_info_list:
                logging.error(f"Collision item: {file_info2}")
            state_error = True
            break
    return state_error



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=unused-variable
# pylint: disable=unused-import
# pylint: disable=wrong-import-position

import os
from pathlib import Path
import logging
from pytest import (
    LogCaptureFixture,
    CaptureFixture,
    fail,
    raises,
    Pytester,
    FixtureRequest,
    Config,
    RunResult,
    ExitCode,
)
import pytest
from atbu.tools.backup.constants import DEFAULT_AES_KEY_BIT_LENGTH

from atbu.tools.backup.exception import (
    PasswordAuthenticationFailure,
)
from atbu.tools.backup.credentials import (
    CredentialByteArray,
    Credential,
    CredentialAesKey,
)

LOGGER = logging.getLogger(__name__)

# import pdb; pdb.set_trace()
# import pdb; pdb.set_trace()

SIMPLE_SECRET = "secret-1234567890!"
SIMPLE_PASSWORD = "1234567890abcdefghij$"


def setup_module(module):  # pylint: disable=unused-argument
    pass


def teardown_module(module):  # pylint: disable=unused-argument
    pass


def test_credential():
    cred1 = Credential(
        the_key=CredentialByteArray.create_from_string(SIMPLE_SECRET),
        password=CredentialByteArray.create_from_string(SIMPLE_PASSWORD),
    )
    cred1.encrypt_key()
    enc_key_bytes = cred1.get_enc_key_material_as_bytes()

    cred2 = Credential.create_credential_from_bytes(enc_key_bytes)
    assert isinstance(cred2, Credential)
    assert cred1.the_key != cred2.the_key
    cred2.set(password=SIMPLE_PASSWORD)
    cred2.decrypt_key()
    assert cred1.the_key == cred2.the_key

    cred2b = Credential.create_credential_from_bytes(
        cred2.get_unenc_key_material_as_bytes()
    )
    assert isinstance(cred2b, Credential)
    assert cred2.the_key == cred2b.the_key

    cred3 = Credential.create_credential_from_bytes(enc_key_bytes)
    assert isinstance(cred3, Credential)
    assert cred1.the_key != cred3.the_key
    cred3.set(password=SIMPLE_PASSWORD + "-incorrect")
    with pytest.raises(PasswordAuthenticationFailure):
        cred3.decrypt_key()
    assert cred1.the_key != cred3.the_key

    cred4 = Credential(
        password=CredentialByteArray.create_from_string(SIMPLE_PASSWORD),
        iv=cred1.iv,
        password_auth_hash=cred1.password_auth_hash,
        salt=cred1.salt,
        encrypted_key=cred1.encrypted_key,
    )
    assert cred1.the_key != cred4.the_key
    cred4.decrypt_key()
    assert cred1.the_key == cred4.the_key


def test_credential_AES():
    the_key = CredentialByteArray(os.urandom(DEFAULT_AES_KEY_BIT_LENGTH // 8))
    cred1 = CredentialAesKey(
        the_key=the_key,
        password=CredentialByteArray.create_from_string(SIMPLE_PASSWORD),
    )
    cred1.encrypt_key()
    enc_key_bytes = cred1.get_enc_key_material_as_bytes()

    cred2 = CredentialAesKey.create_credential_from_bytes(enc_key_bytes)
    assert isinstance(cred2, CredentialAesKey)
    assert cred1.the_key != cred2.the_key
    cred2.set(password=SIMPLE_PASSWORD)
    cred2.decrypt_key()
    assert cred1.the_key == cred2.the_key

    cred2b = CredentialAesKey.create_credential_from_bytes(
        cred2.get_unenc_key_material_as_bytes()
    )
    assert isinstance(cred2b, CredentialAesKey)
    assert cred2.the_key == cred2b.the_key

    cred3 = CredentialAesKey.create_credential_from_bytes(enc_key_bytes)
    assert isinstance(cred3, CredentialAesKey)
    assert cred1.the_key != cred3.the_key
    cred3.set(password=SIMPLE_PASSWORD + "-incorrect")
    with pytest.raises(PasswordAuthenticationFailure):
        cred3.decrypt_key()
    assert cred1.the_key != cred3.the_key

    cred4 = Credential(
        password=CredentialByteArray.create_from_string(SIMPLE_PASSWORD),
        iv=cred1.iv,
        password_auth_hash=cred1.password_auth_hash,
        salt=cred1.salt,
        encrypted_key=cred1.encrypted_key,
    )
    assert cred1.the_key != cred4.the_key
    cred4.decrypt_key()
    assert cred1.the_key == cred4.the_key


def test_credential_AES_create_key():
    cred1 = CredentialAesKey()
    cred1.create_key()
    cred1.set(password=SIMPLE_PASSWORD)
    cred1.encrypt_key()
    enc_key_bytes = cred1.get_enc_key_material_as_bytes()

    cred2 = CredentialAesKey.create_credential_from_bytes(enc_key_bytes)
    assert isinstance(cred2, CredentialAesKey)
    assert cred1.the_key != cred2.the_key
    cred2.set(password=SIMPLE_PASSWORD)
    cred2.decrypt_key()
    assert cred1.the_key == cred2.the_key

    cred3 = CredentialAesKey()
    cred3b = CredentialAesKey()
    cred3.create_key()
    cred3b.create_key()
    assert cred1.the_key != cred3.the_key
    assert cred3.the_key != cred3b.the_key



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Storage interface for local file system.
"""

from glob import iglob
from pathlib import Path
from typing import Iterator, Union
from uuid import uuid4

from atbu.common.exception import (
    exc_to_string,
    ObjectDoesNotExistError,
    ContainerAlreadyExistsError,
)

from .base import (
    AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR,
    StorageContainerInterface,
    StorageInterface,
    StorageObjectInterface,
)
from ..constants import *
from ..config import *
from ...persisted_info.file_info import FileInformation


def get_filesystem_storage_path(root: Union[Path, str], object_name: Union[Path, str]):
    if isinstance(root, str):
        root = Path(root)
    if isinstance(object_name, str):
        object_name = Path(object_name)
    # Path objects have parent . if just a single component.
    if object_name.parent != Path("."):
        raise ValueError("Object name is not just a name.")
    if len(str(object_name)) < 2:
        return root / object_name
    return root / str(object_name)[:2] / object_name


class FileSystemStorageObjectInterface(StorageObjectInterface):
    def __init__(self, object_path: Path):
        super().__init__()
        self.object_path: Path = convert_to_pathlib_path(object_path)
        self.file_info = FileInformation(path=str(object_path))

    @property
    def size(self):
        size_in_bytes = self.file_info.size_in_bytes
        return size_in_bytes

    @property
    def hash(self):
        digest = self.file_info.primary_digest
        return digest

    @property
    def name(self):
        return self.object_path.name


class FileSystemStorageContainerInterface(StorageContainerInterface):
    def __init__(self, container_root_path: Path):
        super().__init__()
        self.container_root_path = container_root_path

    @property
    def name(self) -> str:
        return self.container_root_path

    def get_object(self, object_name: str):
        desired_object_path = get_filesystem_storage_path(
            root=self.container_root_path, object_name=object_name
        )
        if not desired_object_path.is_file():
            raise ObjectDoesNotExistError(
                f"The file {str(desired_object_path)} does not exist or is not a file."
            )
        try:
            return FileSystemStorageObjectInterface(object_path=desired_object_path)
        except Exception as ex:
            logging.error(
                f"Unexpected error trying to access the file: "
                f"{str(desired_object_path)} {exc_to_string(ex)}"
            )
            raise

    def delete_object(self, object_name: str):
        desired_object_path = get_filesystem_storage_path(
            root=self.container_root_path, object_name=object_name
        )
        try:
            desired_object_path.unlink()
        except FileNotFoundError as ex:
            raise ObjectDoesNotExistError(
                f"The object {object_name} does not exist or was not found. {exc_to_string(ex)}"
            ).with_traceback(ex.__traceback__) from ex

    def list_objects(self, prefix: str = None) -> list[StorageObjectInterface]:
        result: list[StorageObjectInterface] = []
        if prefix:
            prefix = os.path.normcase(prefix)
        for p in iglob(os.path.join(self.container_root_path, "**"), recursive=True):
            if prefix:
                just_name = os.path.normcase(os.path.split(p)[1])
                if not just_name.startswith(prefix):
                    continue
            result.append(FileSystemStorageObjectInterface(object_path=p))
        return result


class FileSystemStorageInterface(StorageInterface):
    def __init__(self, storage_def):
        super().__init__()
        self.storage_def: dict = storage_def
        self.provider_id = storage_def[CONFIG_VALUE_NAME_PROVIDER]
        self.backup_root = Path(storage_def[CONFIG_VALUE_NAME_CONTAINER])

    def get_container(self, container_name: str) -> StorageContainerInterface:
        container_name_path = Path(container_name)
        if self.backup_root != container_name_path:
            raise ValueError(
                f"This file system interface only supports "
                f"accessing one container: container={self.backup_root}"
            )
        return FileSystemStorageContainerInterface(self.backup_root)

    def create_container(self, container_name: str) -> StorageContainerInterface:
        if container_name[-1] == AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR:
            base_name = container_name[:-1]
            max_attempts = 1000
            while max_attempts > 0:
                max_attempts -= 1
                candidate = Path(f"{base_name}-{str(uuid4())}")
                if not candidate.exists():
                    candidate.mkdir(parents=True, exist_ok=True)
                    return self.get_container(container_name=candidate)
        else:
            candidate = Path(container_name)
            if candidate.exists():
                raise ContainerAlreadyExistsError(
                    f"FileSystemStorageInterface: The container already exists: {str(candidate)}"
                )
            candidate.mkdir(parents=True, exist_ok=True)
            return self.get_container(container_name=str(candidate))

    def delete_container(self, container_name: str):
        # TBD
        raise NotImplementedError()

    def upload_stream_to_object(
        self,
        container: StorageContainerInterface,
        object_name: str,
        stream,
        source_path: str,
    ):
        if not isinstance(container, FileSystemStorageContainerInterface):
            raise ValueError(
                f"The container is not an instance of FileSystemStorageContainerInterface"
            )
        fsc: FileSystemStorageContainerInterface = container
        if self.backup_root != fsc.container_root_path:
            raise ValueError(
                f"The container does not belong to this interface: "
                f"caller={fsc.container_root_path} this={fsc.container_root_path}"
            )
        storage_path = get_filesystem_storage_path(
            root=self.backup_root, object_name=object_name
        )
        storage_path.parent.mkdir(parents=True, exist_ok=True)
        with open(str(storage_path), "wb") as storage_file:
            for chunk in stream:
                storage_file.write(chunk)

    def download_object_as_stream(
        self, storage_object: FileSystemStorageObjectInterface, chunk_size: int
    ) -> Iterator[bytes]:
        storage_path = get_filesystem_storage_path(
            root=self.backup_root, object_name=storage_object.name
        )
        with open(str(storage_path), "rb") as storage_file:
            while True:
                b = storage_file.read(chunk_size)
                yield b
                if len(b) == 0:
                    break

    @property
    def retry_exceptions(self) -> tuple:
        return ()



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Base for ATBU storage interface abstraction, both decoupling ATBU from
other APIs, and providing ATBU itself a common storage interface for its
focused needs.
"""

from abc import ABC, abstractmethod
from socket import gaierror
from http.client import NotConnected, ImproperConnectionState
from typing import Iterator

from atbu.common.exception import RetryLimitReached

from ..constants import *
from ..config import *
from ..exception import BackupException

CHUNK_SIZE_5MB = 5 * 1024 * 1024
CHUNK_SIZE_50MB = 50 * 1024 * 1024
DEFAULT_CHUNK_UPLOAD_SIZE = CHUNK_SIZE_5MB
DEFAULT_CHUNK_DOWNLOAD_SIZE = CHUNK_SIZE_50MB
DEFAULT_MAX_SIMULTANEOUS_FILE_BACKUPS = 5

DEFAULT_RETRY_EXCEPTIONS = (
    OSError,
    gaierror,
    NotConnected,
    ImproperConnectionState,
    RetryLimitReached,
)

AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR = "*"


class StorageObjectInterface(ABC):
    def __init__(self):
        pass

    @property
    @abstractmethod
    def size(self):
        return None

    @property
    @abstractmethod
    def hash(self):
        return None

    @property
    @abstractmethod
    def name(self):
        return None


class StorageContainerInterface(ABC):
    def __init__(self):
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_object(self, object_name: str) -> StorageObjectInterface:
        pass

    @abstractmethod
    def delete_object(self, object_name: str):
        pass

    @abstractmethod
    def list_objects(self, prefix: str = None) -> list[StorageObjectInterface]:
        pass


class StorageInterface(ABC):
    def __init__(self):
        self._retry_exceptions = list(DEFAULT_RETRY_EXCEPTIONS)

    @property
    def upload_chunk_size(self):
        return DEFAULT_CHUNK_UPLOAD_SIZE

    @property
    def download_chunk_size(self):
        return DEFAULT_CHUNK_DOWNLOAD_SIZE

    @abstractmethod
    def get_container(self, container_name: str) -> StorageContainerInterface:
        pass

    @abstractmethod
    def create_container(self, container_name: str) -> StorageContainerInterface:
        pass

    @abstractmethod
    def delete_container(self, container_name: str):
        pass

    @abstractmethod
    def upload_stream_to_object(
        self,
        container: StorageContainerInterface,
        object_name: str,
        stream,
        source_path: str,
    ):
        pass

    @abstractmethod
    def download_object_as_stream(
        self, storage_object: StorageObjectInterface, chunk_size: int
    ) -> Iterator[bytes]:
        pass

    @property
    def retry_exceptions(self):
        return self._retry_exceptions


class StorageInterfaceFactory:
    def __init__(self, storage_def_dict):
        self.storage_def_dict = storage_def_dict

    @property
    def interface_type(self):
        return self.storage_def_dict[CONFIG_VALUE_NAME_INTERFACE_TYPE]

    @property
    def provider_name(self):
        return self.storage_def_dict[CONFIG_VALUE_NAME_PROVIDER]

    def create_storage_interface(self) -> StorageInterface:
        # pylint: disable=import-outside-toplevel
        desired_interface = self.storage_def_dict[CONFIG_VALUE_NAME_INTERFACE_TYPE]
        if desired_interface == CONFIG_INTERFACE_TYPE_LIBCLOUD:
            from .libcloud import LibCloudStorageInterface

            return LibCloudStorageInterface(storage_def=self.storage_def_dict)
        elif desired_interface == CONFIG_INTERFACE_TYPE_AZURE:
            from .azure import AzureStorageInterface

            return AzureStorageInterface(storage_def=self.storage_def_dict)
        elif desired_interface == CONFIG_INTERFACE_TYPE_GOOGLE:
            from .google import GoogleStorageInterface

            return GoogleStorageInterface(storage_def=self.storage_def_dict)
        elif desired_interface == CONFIG_INTERFACE_TYPE_FILESYSTEM:
            from .filesystem import FileSystemStorageInterface

            return FileSystemStorageInterface(storage_def=self.storage_def_dict)
        else:
            raise BackupException(f"Unknown interface type: {desired_interface}")

    @staticmethod
    def create_factory_from_storage_def_name(storage_def_name: str):
        atbu_cfg, _, _ = AtbuConfig.access_cloud_storage_config(
            storage_def_name=storage_def_name,
            must_exist=True,
            create_if_not_exist=False,
        )
        storage_def_dict = atbu_cfg.get_storage_def_with_resolved_secrets_deep_copy(
            storage_def_name=storage_def_name
        )
        return StorageInterfaceFactory(storage_def_dict=storage_def_dict)



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""ChunkSizeFileReader simplifies enforcing reading from a file
in byte chunks of a specified size. This is important for particular
APIs such as libcloud which expect to see data in specific chunk
sizes until the last chunk.
"""

from io import RawIOBase
from typing import Callable

from atbu.common.exception import AlreadyUsedError
from atbu.common.aes_cbc import AesCbcPaddingEncryptor

CHUNK_READER_CB_INPUT_BYTES = "input-bytes"
CHUNK_READER_CB_INPUT_BYTES_MANUAL_APPEND = "input-bytes-append"
CHUNK_READER_CB_CIPHERTEXT = "output-ciphertext"


class ChunkSizeFileReader:
    """Allows reading a file without encryption in chunks of the specified size.
    This is provided to allow consistent interfaces for reading chunk size
    bytes of a file as its encrypted as well as in cases where it is not encrypted.

    Instances can be used as an iterator or by manually calling methods.
    """

    def __init__(
        self,
        chunk_size,
        path,
        fileobj: RawIOBase,
        read_without_size: bool,
        user_func: Callable[[str, bytes], None],
    ):
        self._chunk_size = chunk_size
        self.path = path
        self._file = fileobj
        self.read_without_size = read_without_size
        self._eof_detected = False
        self._is_used = False
        self._user_func = user_func
        self._pending_output = bytes()

    def __iter__(self):
        return self

    def __next__(self):
        if self._eof_detected:
            raise StopIteration(f"No more chunks from file: {self.path}")
        chunk = self.read_chunk()
        return chunk

    def _checkused(self):
        if self._is_used:
            raise AlreadyUsedError(f"This instance cannot be used more than once.")

    def _call_user_func(self, what, data):
        if not self._user_func:
            return None
        return self._user_func(what, data)

    def open(self):
        self._checkused()
        if self._file:
            return
        self._file = open(self.path, "rb")

    def close(self):
        if not self._file:
            return
        self._checkused()
        self._file.close()
        self._file = None
        self._is_used = True

    def queue_data(self, bytes_to_queue, do_not_encrypt: bool = True):
        if not do_not_encrypt:
            raise ValueError(
                f"ChunkSizeFileReader: Encryption is not active, "
                f"do_not_encrypt=False is disallowed."
            )
        self._checkused()
        if self._eof_detected:
            raise EOFError(
                f"EOF empty bytes already returned to indicate EOF, cannot call read again."
            )
        if not self._file:
            raise Exception(f"The file is not open.")
        if len(bytes_to_queue) == 0:
            raise ValueError(f"The data must contain bytes to write.")
        if len(self._pending_output) == 0:
            # If caller passes bytearray, things work fine here but certain APIs
            # reading chunks choke on bytearray instead of bytes. Regardless,
            # our internal type is be consistent, not defined by caller's insertion
            # using bytearray. At this time insert using write_data is minimal and
            # read_chunk attempts to align on chunk bytes reads to mitigate copying.
            self._pending_output = bytes(bytes_to_queue)
        else:
            self._pending_output += bytes_to_queue
        self._call_user_func(CHUNK_READER_CB_INPUT_BYTES, bytes_to_queue)

    def read_chunk(self):
        self._checkused()
        if self._eof_detected:
            raise EOFError(
                f"EOF empty bytes already returned to indicate EOF, cannot call read again."
            )
        if not self._file:
            raise Exception(
                f"This instance must be entered and not closed in order to properly exit."
            )
        while len(self._pending_output) < self._chunk_size:
            size_to_read = None
            if not self.read_without_size:
                size_to_read = self._chunk_size - len(self._pending_output)
            new_file_bytes = self._file.read(size_to_read)
            if len(new_file_bytes) == 0:
                break
            self._call_user_func(CHUNK_READER_CB_INPUT_BYTES, new_file_bytes)
            if len(self._pending_output) == 0:
                self._pending_output = new_file_bytes
            else:
                self._pending_output += new_file_bytes

        if len(self._pending_output) <= self._chunk_size:
            bytes_to_return = self._pending_output
            self._pending_output = bytes()
        else:
            bytes_to_return = self._pending_output[: self._chunk_size]
            self._pending_output = self._pending_output[self._chunk_size :]
        self._eof_detected = len(bytes_to_return) == 0
        return bytes_to_return

    def read(self, size):
        if size != self._chunk_size:
            raise Exception(
                f"You are only allowed to read chunk sizes of "
                f"{self._chunk_size} bytes from this instance."
            )
        return self.read_chunk()

    def __enter__(self):
        self._checkused()
        if self._file:
            return self
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._checkused()
        if not self._file:
            raise Exception(
                f"This instance must be entered and not closed in order to properly exit."
            )
        self.close()
        return False


class ChunkSizeEncryptorReader(ChunkSizeFileReader):
    """Read ciphertext of file encrypted with AesCbcPaddingEncryptor in specified chunk size.
    The read_chunk() method will always return one chunk until EOF at which time the remaining
    ciphertext is returned after which zero bytes. This class effectively hides the details of
    any internal overhead or buffering of AesCbcPaddingEncryptor, allowing users of this class
    to see clean chunk sizes of ciphertext until EOF.
    """

    def __init__(
        self,
        chunk_size,
        path,
        fileobj: RawIOBase,
        read_without_size: bool,
        encryptor: AesCbcPaddingEncryptor,
        user_func: Callable[[str, bytes], None],
    ):
        super().__init__(
            chunk_size=chunk_size,
            path=path,
            fileobj=fileobj,
            read_without_size=read_without_size,
            user_func=user_func,
        )
        if not isinstance(encryptor, AesCbcPaddingEncryptor):
            raise ValueError(
                f"The encryptor must be an instance of AesCbcPaddingEncryptor."
            )
        self.encryptor = encryptor

    @property
    def is_encryption_finished(self):
        return self.encryptor.finalize

    def queue_data(self, bytes_to_queue, do_not_encrypt: bool = False):
        """Queues data to be read as part of a chunk during a subsequent call to read_chunk.
        This method offers a way to "insert" data into the stream obtained via read_chunk.

        Specify do_not_encrypt=True to insert information which should not be encrypted. This would
        usually only be used at the start of writing to include unecrypted prefix information in a
        storage blob, information such as backup format versioning and option bits, IVs, etc.
        """
        self._checkused()
        if self._eof_detected:
            raise EOFError(
                f"EOF empty bytes already returned to indicate EOF, cannot call read again."
            )
        if not self._file:
            raise Exception(f"The file is not open.")
        if len(bytes_to_queue) == 0:
            raise ValueError(f"The data must contain bytes to write.")
        self._call_user_func(CHUNK_READER_CB_INPUT_BYTES_MANUAL_APPEND, bytes_to_queue)
        if do_not_encrypt:
            if len(self._pending_output) == 0:
                self._pending_output = bytes_to_queue
            else:
                self._pending_output += bytes_to_queue
        else:
            new_cipher_text = self.encryptor.update(input_bytes=bytes_to_queue)
            if len(new_cipher_text) > 0:
                self._call_user_func(CHUNK_READER_CB_CIPHERTEXT, new_cipher_text)
                if len(self._pending_output) == 0:
                    self._pending_output = new_cipher_text
                else:
                    self._pending_output += new_cipher_text

    def read_chunk(self):
        self._checkused()
        if self._eof_detected:
            raise EOFError(
                f"EOF empty bytes already returned to indicate EOF, cannot call read again."
            )
        if not self._file:
            raise Exception(f"The file is not open.")

        if self.encryptor.is_finalized:
            if len(self._pending_output) <= self._chunk_size:
                ciphertext_chunk_to_return = self._pending_output
                self._pending_output = bytes()
            else:
                ciphertext_chunk_to_return = self._pending_output[: self._chunk_size]
                self._pending_output = self._pending_output[self._chunk_size :]
            self._eof_detected = len(ciphertext_chunk_to_return) == 0
            return ciphertext_chunk_to_return

        while len(self._pending_output) < self._chunk_size:
            size_to_read = None
            if not self.read_without_size:
                size_to_read = (
                    self._chunk_size
                    - len(self._pending_output)
                    + (self.encryptor.BLOCK_SIZE * 3)
                )
            plaintext_bytes = self._file.read(size_to_read)
            if len(plaintext_bytes) == 0:
                new_cipher_text = self.encryptor.finalize()
            else:
                new_cipher_text = self.encryptor.update(input_bytes=plaintext_bytes)
                self._call_user_func(CHUNK_READER_CB_INPUT_BYTES, plaintext_bytes)
            self._pending_output += new_cipher_text
            if len(new_cipher_text) > 0:
                self._call_user_func(CHUNK_READER_CB_CIPHERTEXT, new_cipher_text)
            if len(plaintext_bytes) == 0:
                break

        if len(self._pending_output) <= self._chunk_size:
            ciphertext_chunk_to_return = self._pending_output
            self._pending_output = bytes()
        else:
            ciphertext_chunk_to_return = self._pending_output[: self._chunk_size]
            self._pending_output = self._pending_output[self._chunk_size :]

        self._eof_detected = len(ciphertext_chunk_to_return) == 0
        return ciphertext_chunk_to_return


def open_chunk_reader(
    chunk_size,
    path,
    fileobj: RawIOBase,
    read_without_size: bool,
    encryptor: AesCbcPaddingEncryptor,
    user_func: Callable[[str, bytes], None],
):
    if encryptor:
        return ChunkSizeEncryptorReader(
            chunk_size=chunk_size,
            path=path,
            fileobj=fileobj,
            read_without_size=read_without_size,
            encryptor=encryptor,
            user_func=user_func,
        )
    return ChunkSizeFileReader(
        chunk_size=chunk_size,
        path=path,
        fileobj=fileobj,
        read_without_size=read_without_size,
        user_func=user_func,
    )



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""ATBU Configuration-related migration helper classes/functions.
"""

import base64
from dataclasses import dataclass
import keyring

from atbu.common.util_helpers import (
    is_valid_base64_string,
)

# Import module to mock credentials.* once directly.
import atbu.tools.backup.credentials as credentials
from .constants import (
    CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
    CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
    CONFIG_PASSWORD_KIND_ACTUAL,
    CRED_SECRET_KIND_ENCRYPTION,
    CRED_SECRET_KIND_STORAGE,
)
from .exception import (
    ConfigMigrationError,
    CredentialNotFoundError,
    InvalidBase64StringError,
    PasswordAuthenticationFailure,
    CredentialInvalid,
    CredentialSecretDerivationError,
)
from .credentials import (
    Credential,
    CredentialByteArray,
    CredentialAesKey,
    DescribedCredential,
)
from .storage_def_credentials import StorageDefCredentialSet


@dataclass
class Version_001:
    AT_PREFIX = "at"
    AT_PREFIX_U = "AT"
    ATBU_ACRONYM = f"{AT_PREFIX}bu"
    ATBU_ACRONUM_U = f"{AT_PREFIX_U}BU"

    CONFIG_SECTION_KEYRING_MAPPING = "keyring-mapping"
    CONFIG_SECTION_KEYRING_MAPPING_ENCRYPTION_KEY = "encryption-key"
    CONFIG_SECTION_KEYRING_MAPPING_DRIVER_SECRET = "driver-secret"

    CONFIG_VALUE_NAME_KEYRING_SERVICE = "service"
    CONFIG_VALUE_NAME_KEYRING_USERNAME = "username"

    CONFIG_KEY_VALUE_KEYRING_INDIRECTION = "keyring"

    CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION = f"{ATBU_ACRONUM_U}-backup-enc-key"
    CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD = f"{ATBU_ACRONUM_U}-storage-password"

    CRED_KEY_TYPE_STORAGE = "storage-secret"
    CRED_KEY_TYPE_ENCRYPTION = "encryption-key"

    CONFIG_PASSWORD_TYPE = "password_type"
    CONFIG_PASSWORD_TYPE_ACTUAL = "actual"
    CONFIG_PASSWORD_TYPE_FILENAME = "filename"
    CONFIG_PASSWORD_TYPE_ENVVAR = "envvar"
    PASSWORD_TYPE_CHAR_TO_TYPE = {
        i[0]: i
        for i in [
            CONFIG_PASSWORD_TYPE_ACTUAL,
            CONFIG_PASSWORD_TYPE_FILENAME,
            CONFIG_PASSWORD_TYPE_ENVVAR,
        ]
    }

    CONFIG_SECTION_DRIVER = "driver"
    CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET = "secret"

    CONFIG_SECTION_ENCRYPTION = "encryption"
    CONFIG_VALUE_NAME_ENCRYPTION_KEY = "key"

    CONFIG_SECTION_STORAGE_DEFINITIONS = "storage-definitions"


def set_password_to_keyring_001(
    service_name: str,
    username: str,
    password_type: str,
    password_bytes: CredentialByteArray,
    password_is_base64: bool = False,
) -> None:
    if not service_name:
        raise CredentialInvalid(
            f"Cannot store credential. The supplied service name must be a non-empty string."
        )

    if not username:
        raise CredentialInvalid(
            f"Cannot store credential. The supplied user name must be a non-empty string."
        )

    if not password_type:
        raise CredentialInvalid(
            f"Cannot store credential. The supplied password type must be a non-empty string."
        )

    password_type_code = password_type[0]
    if password_type_code not in Version_001.PASSWORD_TYPE_CHAR_TO_TYPE:
        raise CredentialInvalid(
            f"Cannot store credential. The supplied password type '{password_type}' is invalid."
        )

    if not password_bytes:
        raise CredentialInvalid(
            f"Cannot store credential. The password was not supplied."
        )

    if not password_is_base64:
        password_bytes = CredentialByteArray(base64.b64encode(password_bytes))

    if not is_valid_base64_string(str_to_check=password_bytes):
        raise InvalidBase64StringError(
            f"Expected password_bytes to be a base64-encoded string."
        )

    keyring.set_password(
        service_name=service_name,
        username=username,
        password=f"{password_type_code}:{str(password_bytes, 'utf-8')}",
    )


def _get_base64_password_from_keyring(
    service_name: str,
    username: str,
) -> tuple[str, str, CredentialByteArray]:
    raw_password = keyring.get_password(service_name=service_name, username=username)
    if raw_password is None:
        raise CredentialNotFoundError(
            f"The credential was not found in the keyring: "
            f"service_name={service_name} username={username}"
        )
    password_parts = raw_password.split(":", maxsplit=1)
    if len(password_parts) != 2:
        raise CredentialInvalid(
            f"The stored credential is not in the expected two-part format."
        )
    password_type_code = password_parts[0]
    cba_password_base64 = CredentialByteArray(password_parts[1].encode("utf-8"))

    if len(password_type_code) == 0:
        raise CredentialInvalid(
            f"The stored credential is not in the expected format. "
            f"The password type code is not present. "
            f"service_name={service_name} username={username}"
        )

    if len(cba_password_base64) == 0:
        raise CredentialInvalid(
            f"The stored credential is not in the expected format. "
            f"The password is not present. "
            f"service_name={service_name} username={username}"
        )

    if not is_valid_base64_string(cba_password_base64):
        raise InvalidBase64StringError(
            f"Expected stored credential to be in base64 format."
        )

    if password_type_code not in Version_001.PASSWORD_TYPE_CHAR_TO_TYPE:
        raise CredentialInvalid(
            f"The stored credential is not in the expected format. "
            f"The password type code '{password_type_code}' is unknown. "
            f"service_name={service_name} username={username}"
        )

    password_type = Version_001.PASSWORD_TYPE_CHAR_TO_TYPE[password_type_code]

    return (password_type_code, password_type, cba_password_base64)


def _unlock_credential_001(credential: CredentialAesKey):
    """Unlock a CredentialAesKey instance which is to make its private key/secret
    available in the clear. If the private key is password-protected, ask
    for a password. The result of calling this function is an unlocked
    credential or an exception.

        Exceptions:
            PasswordAuthenticationFailure: Too many attempts with incorrect password.
            CredentialSecretDerivationError: Something unexpected indicates the key
            is not available despite having been decrypted (likely a program bug or
            something esoteric such as system corruption).
    """
    if not credential.is_private_key_possible:
        raise CredentialInvalid(
            f"The private key is not available. "
            f"The credential is invalid or corrupt."
        )
    if not credential.is_private_key_ready:
        if not credential.is_password_required:
            credential.decrypt_key()
        else:
            attempts = 5
            while True:
                password = credentials.prompt_for_password_with_yubikey_opt(
                    prompt="Enter the password for this backup:"
                )
                credential.set(password=password)
                try:
                    attempts -= 1
                    credential.decrypt_key()
                    break
                except PasswordAuthenticationFailure:
                    if attempts > 0:
                        print(f"The password appears to be invalid, try again.")
                    else:
                        print(f"Still incorrect.")
                        raise
        if not credential.is_private_key_ready:
            raise CredentialSecretDerivationError(
                "Unexpected failure, canonot access the credential."
            )


def _get_password_from_keyring(
    service_name: str,
    username: str,
) -> tuple[str, str, CredentialByteArray]:
    (
        password_type_code,
        password_type,
        cba_password_base64,
    ) = _get_base64_password_from_keyring(
        service_name=service_name,
        username=username,
    )

    cba_password = CredentialByteArray(base64.b64decode(cba_password_base64))

    if (
        username == Version_001.CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION
        and password_type == Version_001.CONFIG_PASSWORD_TYPE_ACTUAL
    ):
        credential = CredentialAesKey.create_credential_from_bytes(
            cred_bytes=cba_password
        )
        _unlock_credential_001(credential=credential)
        # The actual 0.01 code only propagates the key but migration requires
        # the CredentialAesKey instance.
        cba_password = credential
        # cba_password = credential.the_key
        # del credential

    return (password_type_code, password_type, cba_password)


def _resolve_keyring_secrets(
    storage_def_dict: dict,
):
    # Access 'keyring-mapping' section.
    keyring_section: dict = storage_def_dict.get(
        Version_001.CONFIG_SECTION_KEYRING_MAPPING
    )
    if not keyring_section:
        return

    # For each 'keyring-mapping' item:
    #    item's name:
    #       encoded with string such as 'driver-secret' to indicate mapping
    #       relates from secret to { ... "driver": { ... "secret": <secret_access_key> ... } ... }
    #    items's value:
    #       indicates keyring service/username with secret.
    affected_config_path: str
    for affected_config_path, keyring_lookup_info in keyring_section.items():
        # Split item's name to get the storage definition path parts.
        affected_config_path_parts = affected_config_path.split("-")

        # Get the service/user names.
        service_name = keyring_lookup_info[
            Version_001.CONFIG_VALUE_NAME_KEYRING_SERVICE
        ]
        username = keyring_lookup_info[Version_001.CONFIG_VALUE_NAME_KEYRING_USERNAME]

        # Walk from top of config_section down to the second-to-last section which is
        # where the secret goes.
        affected_section = storage_def_dict
        for section_part in affected_config_path_parts[:-1]:
            affected_section = affected_section[section_part]

        # Given service/user names, get the secret.
        (_, password_type, cba_password) = _get_password_from_keyring(
            service_name=service_name,
            username=username,
        )

        affected_section[affected_config_path_parts[-1]] = cba_password
        affected_section[Version_001.CONFIG_PASSWORD_TYPE] = password_type


@dataclass
class StorageDefMigrationItem_001:
    storage_def_name: str
    storage_def_dict: dict
    cred_set: StorageDefCredentialSet


def prepare_for_storage_def_migration_001(
    cfg: dict,
) -> list[StorageDefMigrationItem_001]:
    """For each storage definition, return a StorageDefMigrationItem_001 containing
    the storage def's credentials in a StorageDefCredentialSet. The set will be
    unprotected and unsaved upon return (i.e., not written yet). For each storage
    definition in the supplied configuration, update the dictionary to remove the
    old keyring mapping. When the caller writes the cred set, the dictionary will
    be updated further.

    Args:
        cfg (dict): The v0.01 ATBU configuration to convert to v0.02.

    Returns:
        list[StorageDefMigrationItem_001]: One migration item for each storage
        definition found.
    """
    migration_items = list[StorageDefMigrationItem_001]()
    storage_definitions = cfg[Version_001.CONFIG_SECTION_STORAGE_DEFINITIONS]
    storage_def_name: str
    storage_def_dict: dict
    for storage_def_name, storage_def_dict in storage_definitions.items():
        print(f"Processing storage definition: {storage_def_name}...")
        _resolve_keyring_secrets(storage_def_dict=storage_def_dict)
        if storage_def_dict.get(Version_001.CONFIG_SECTION_KEYRING_MAPPING) is not None:
            del storage_def_dict[Version_001.CONFIG_SECTION_KEYRING_MAPPING]
        cred_storage = None
        cred_storage_type = None
        cred_encryption = None
        storage_section = storage_def_dict.get(Version_001.CONFIG_SECTION_DRIVER)
        if isinstance(storage_section, dict) and len(storage_section) != 0:
            storage_secret = storage_section.get(
                Version_001.CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET
            )
            if not isinstance(storage_secret, CredentialByteArray):
                raise ConfigMigrationError(
                    f"Expected storage secret to be CredentialByteArray but got "
                    f"'{type(storage_secret)}'."
                )
            cred_storage = Credential(the_key=storage_secret)
            cred_storage_type = storage_section.get(Version_001.CONFIG_PASSWORD_TYPE)
            if not isinstance(cred_storage_type, str) or len(cred_storage_type) == 0:
                raise ConfigMigrationError(
                    f"Expected non-zero len str cred_storage_type: "
                    f"type={type(cred_storage_type)} "
                    f"len={len(cred_storage_type)}"
                )
        else:
            print(f"No cloud storage section found for '{storage_def_name}'.")
        encryption_section = storage_def_dict.get(Version_001.CONFIG_SECTION_ENCRYPTION)
        if isinstance(encryption_section, dict):
            cred_encryption = encryption_section.get(
                Version_001.CONFIG_VALUE_NAME_ENCRYPTION_KEY
            )
            if not isinstance(cred_encryption, CredentialAesKey):
                raise ConfigMigrationError(
                    f"Expected CredentialAesKey but got {type(cred_encryption)}"
                )
            if cred_storage is not None and cred_encryption.is_password_protected:
                if (
                    not isinstance(cred_encryption.password, CredentialByteArray)
                    or len(cred_encryption.password) == 0
                ):
                    raise ConfigMigrationError(
                        f"Password not found for password-protected encryption key: "
                        f"type={type(cred_encryption.password)} "
                        f"len={len(cred_encryption.password)}"
                    )
                cred_storage.prepare_for_new_password()
                cred_storage.set(password=CredentialByteArray(cred_encryption.password))
        cred_set = StorageDefCredentialSet(
            storage_def_name=storage_def_name,
            storage_def_dict=storage_def_dict,
        )
        if cred_encryption is not None:
            desc_cred_encryption = DescribedCredential(
                credential=cred_encryption,
                config_name=storage_def_name,
                credential_name=CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
                credential_kind=CONFIG_PASSWORD_KIND_ACTUAL,
            )
            cred_set.append(
                desc_cred=desc_cred_encryption,
                affected_config_path_parts=CRED_SECRET_KIND_ENCRYPTION.split("-"),
            )
        if cred_storage is not None:
            desc_cred_storage = DescribedCredential(
                credential=cred_storage,
                config_name=storage_def_name,
                credential_name=CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
                credential_kind=cred_storage_type,
            )
            cred_set.append(
                desc_cred=desc_cred_storage,
                affected_config_path_parts=CRED_SECRET_KIND_STORAGE.split("-"),
            )
        migration_items.append(
            StorageDefMigrationItem_001(
                storage_def_name=storage_def_name,
                storage_def_dict=storage_def_dict,
                cred_set=cred_set,
            )
        )
    if len(storage_definitions) != len(migration_items):
        raise ConfigMigrationError(
            f"Mismatched storage definition count: "
            f"config={len(storage_definitions)} "
            f"migration_items={len(migration_items)}."
        )
    return migration_items


def upgrade_storage_definitions_from_001_to_002(
    cfg: dict,
):
    migration_items = prepare_for_storage_def_migration_001(cfg=cfg)
    for mi in migration_items:
        mi.cred_set.protect()
        mi.cred_set.save()



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=unused-variable
# pylint: disable=unused-import
# pylint: disable=wrong-import-position

import base64
import os
from pathlib import Path
import logging
from pytest import (
    LogCaptureFixture,
    CaptureFixture,
    fail,
    raises,
    Pytester,
    FixtureRequest,
    Config,
    RunResult,
    ExitCode,
)
import pytest
from atbu.tools.backup.constants import (
    CONFIG_INTERFACE_TYPE_LIBCLOUD,
    CONFIG_KEY_VALUE_KEYRING_INDIRECTION,
    CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
    CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
    CONFIG_PASSWORD_KIND_ACTUAL,
    CONFIG_PASSWORD_TYPE,
    CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY,
    CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET,
    CRED_SECRET_KIND_ENCRYPTION,
    CRED_SECRET_KIND_STORAGE,
)

from atbu.tools.backup.credentials import (
    CredentialByteArray,
    Credential,
    CredentialAesKey,
    DescribedCredential,
)
from atbu.tools.backup.config import AtbuConfig
from atbu.tools.backup.storage_def_credentials import StorageDefCredentialSet

LOGGER = logging.getLogger(__name__)

ATBU_TEST_BACKUP_NAME = "AtbuTestBackup-5b497bb3-c9ef-48a9-af7b-2327fc17fb65"

# import pdb; pdb.set_trace()
# import pdb; pdb.set_trace()

SIMPLE_SECRET = "secret-1234567890!"
SIMPLE_PASSWORD = "1234567890abcdefghij$"


def setup_module(module):  # pylint: disable=unused-argument
    pass


def teardown_module(module):  # pylint: disable=unused-argument
    pass


def add_encryption_credential_to_set(
    storage_def_name: str,
    cred_set: StorageDefCredentialSet,
    credential_password: CredentialByteArray = None,
):
    cred_encryption = CredentialAesKey()
    cred_encryption.create_key()
    desc_cred_encryption = DescribedCredential(
        credential=cred_encryption,
        config_name=storage_def_name,
        credential_name=CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
        credential_kind=CONFIG_PASSWORD_KIND_ACTUAL,
    )
    cred_set.append(
        desc_cred=desc_cred_encryption,
        affected_config_path_parts=CRED_SECRET_KIND_ENCRYPTION.split("-"),
    )
    if credential_password is not None:
        cred_set.get_encryption_desc_cred().credential.set(
            password=CredentialByteArray(credential_password),
        )


def add_storage_credential_to_set(
    storage_def_name: str,
    cred_set: StorageDefCredentialSet,
    storage_secret: str,
    credential_password: CredentialByteArray = None,
):
    cred_storage = Credential(
        the_key=CredentialByteArray(storage_secret.encode("utf-8"))
    )
    desc_cred_encryption = cred_set.get_encryption_desc_cred()
    if desc_cred_encryption is not None:
        cred_encryption = desc_cred_encryption.credential
        if cred_encryption.is_password_protected:
            cred_storage.set(password=CredentialByteArray(cred_encryption.password))
    desc_cred_storage = DescribedCredential(
        credential=cred_storage,
        config_name=storage_def_name,
        credential_name=CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
        credential_kind=CONFIG_PASSWORD_KIND_ACTUAL,
    )
    cred_set.append(
        desc_cred=desc_cred_storage,
        affected_config_path_parts=CRED_SECRET_KIND_STORAGE.split("-"),
    )
    if credential_password is not None:
        cred_set.get_storage_desc_cred().credential.set(
            password=CredentialByteArray(credential_password),
        )


def create_test_storage_def_credential_set(
    storage_def_name: str,
    storage_def_dict: dict,
    include_storage_cred: bool,
    storage_secret: str,
    credential_password: CredentialByteArray = None,
) -> StorageDefCredentialSet:

    cred_set = StorageDefCredentialSet(
        storage_def_name=storage_def_name,
        storage_def_dict=storage_def_dict,
    )

    add_encryption_credential_to_set(
        storage_def_name=storage_def_name,
        cred_set=cred_set,
        credential_password=credential_password,
    )

    if include_storage_cred:
        add_storage_credential_to_set(
            storage_def_name=storage_def_name,
            cred_set=cred_set,
            storage_secret=storage_secret,
            credential_password=credential_password,
        )

    return cred_set


@pytest.mark.parametrize(
    "is_storage_secret_in_config,is_populate_init,is_password_protected",
    [
        pytest.param(
            False,
            False,
            False,
            id="secret_in_config=False,populate_init=False,password_protected=False",
        ),
        pytest.param(
            True,
            False,
            False,
            id="secret_in_config=True,populate_init=False,password_protected=False",
        ),
        pytest.param(
            True,
            True,
            False,
            id="secret_in_config=True,populate_init=True,password_protected=False",
        ),
        pytest.param(
            False,
            False,
            True,
            id="secret_in_config=False,populate_init=False,password_protected=True",
        ),
        pytest.param(
            True,
            False,
            True,
            id="secret_in_config=True,populate_init=False,password_protected=True",
        ),
        pytest.param(
            True,
            True,
            True,
            id="secret_in_config=True,populate_init=True,password_protected=True",
        ),
    ],
)
def test_storage_def_credential_set(
    tmp_path: Path,
    pytester: Pytester,
    is_storage_secret_in_config: bool,
    is_populate_init: bool,
    is_password_protected: bool,
):
    """Manually create a StorageDefCredentialSet, validate operations/persistence.

    This scenario initializes the StorageDefCredentialSet by building the credentials
    manually in code, and not deriving some/all from the configuration file.
    """
    storage_def_name = "test_storage_def_credential_set"
    atbu_cfg, _, _ = AtbuConfig.access_cloud_storage_config(
        storage_def_name=storage_def_name,
        must_exist=False,
        create_if_not_exist=False,
    )
    assert atbu_cfg is None
    atbu_cfg = AtbuConfig.create_cloud_storage_def_config(
        storage_def_name=storage_def_name,
    )
    credential_password = None
    if is_password_protected:
        credential_password = CredentialByteArray("test-cred-pwd$".encode("utf-8"))

    storage_secret = "test-storage-secret"
    if not is_storage_secret_in_config:
        other_kv_pairs = {
            CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY: "test-storage-key",
        }
    else:
        other_kv_pairs = {
            CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY: "test-storage-key",
            CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET: base64.b64encode(
                f"K={storage_secret.encode('utf-8').hex()}".encode("utf-8")
            ).decode("utf-8"),
            CONFIG_PASSWORD_TYPE: "actual",
        }

    atbu_cfg.create_storage_def(
        interface_type=CONFIG_INTERFACE_TYPE_LIBCLOUD,
        provider_id="test-provider-id",
        container="test-container-name",
        other_driver_kv_pairs=other_kv_pairs,
        unique_storage_def_name=storage_def_name,
    )

    if not is_populate_init:
        cred_set = create_test_storage_def_credential_set(
            storage_def_name=storage_def_name,
            storage_def_dict=atbu_cfg.get_storage_def_dict(
                storage_def_name=storage_def_name
            ),
            include_storage_cred=True,
            storage_secret=storage_secret,
            credential_password=credential_password,
        )
    else:
        cred_set = StorageDefCredentialSet(
            storage_def_name=storage_def_name,
            storage_def_dict=atbu_cfg.get_storage_def_dict(
                storage_def_name=storage_def_name
            ),
        )
        cred_set.populate()

        if credential_password is not None:
            cred_set.set_password(
                cred_password=CredentialByteArray(credential_password)
            )

        assert cred_set.get_encryption_desc_cred() is None
        assert cred_set.get_storage_desc_cred() is not None

        add_encryption_credential_to_set(
            storage_def_name=storage_def_name,
            cred_set=cred_set,
            credential_password=credential_password,
        )

    desc_cred_encryption = cred_set.get_encryption_desc_cred()
    desc_cred_storage = cred_set.get_storage_desc_cred()
    cred_set.protect()
    cred_set.save()
    atbu_cfg.save_config_file()

    atbu_cfg2, _, _ = AtbuConfig.access_cloud_storage_config(
        storage_def_name=storage_def_name,
        must_exist=True,
        create_if_not_exist=False,
    )
    storage_def_dict2 = atbu_cfg2.get_storage_def_dict(
        storage_def_name=storage_def_name
    )
    cred_set = StorageDefCredentialSet(
        storage_def_name=storage_def_name,
        storage_def_dict=storage_def_dict2,
    )

    encryption_dict, encryption_key_value_name = cred_set.get_affected_section(
        CRED_SECRET_KIND_ENCRYPTION.split("-")
    )
    storage_dict, storage_secret_value_name = cred_set.get_affected_section(
        CRED_SECRET_KIND_STORAGE.split("-")
    )

    assert (
        encryption_dict[encryption_key_value_name]
        == CONFIG_KEY_VALUE_KEYRING_INDIRECTION
    )
    assert (
        storage_dict[storage_secret_value_name] == CONFIG_KEY_VALUE_KEYRING_INDIRECTION
    )

    cred_set.populate()
    cred_set.unprotect(password=credential_password)

    assert (
        encryption_dict[encryption_key_value_name]
        != CONFIG_KEY_VALUE_KEYRING_INDIRECTION
    )
    assert (
        storage_dict[storage_secret_value_name] != CONFIG_KEY_VALUE_KEYRING_INDIRECTION
    )
    assert (
        encryption_dict[encryption_key_value_name]
        == desc_cred_encryption.credential.the_key
    )
    assert (
        storage_dict[storage_secret_value_name] == desc_cred_storage.credential.the_key
    )

    desc_cred_encryption2 = cred_set.get_encryption_desc_cred()
    desc_cred_storage2 = cred_set.get_storage_desc_cred()
    assert (
        desc_cred_encryption2.credential.the_key
        == desc_cred_encryption.credential.the_key
    )
    assert desc_cred_encryption2.config_name == desc_cred_encryption.config_name
    assert desc_cred_encryption2.credential_name == desc_cred_encryption.credential_name
    assert desc_cred_encryption2.credential_kind == desc_cred_encryption.credential_kind

    assert desc_cred_storage2.credential.the_key == desc_cred_storage.credential.the_key
    assert desc_cred_storage2.config_name == desc_cred_storage.config_name
    assert desc_cred_storage2.credential_name == desc_cred_storage.credential_name
    assert desc_cred_storage2.credential_kind == desc_cred_storage.credential_kind



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=unused-variable
# pylint: disable=unused-import
# pylint: disable=wrong-import-position

import copy
import json
import os
from pathlib import Path
import logging
from uuid import uuid4
from pytest import (
    LogCaptureFixture,
    CaptureFixture,
    fail,
    raises,
    Pytester,
    FixtureRequest,
    Config,
    RunResult,
    ExitCode,
    MonkeyPatch,
)
import pytest

from atbu.tools.backup.config import (
    AtbuConfig,
)
from atbu.tools.backup.constants import (
    ATBU_CONFIG_FILE_VERSION_STRING_CURRENT,
    CONFIG_PASSWORD_KIND_ACTUAL,
    CONFIG_PASSWORD_TYPE,
    CONFIG_SECTION_DRIVER,
    CONFIG_VALUE_NAME_CONTAINER,
    CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY,
    CONFIG_VALUE_NAME_DRIVER_STORAGE_PROJECT,
    CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET,
    CONFIG_VALUE_NAME_INTERFACE_TYPE,
    CONFIG_VALUE_NAME_PROVIDER,
    CONFIG_VALUE_NAME_STORAGE_DEF_UNIQUE_ID,
)
from atbu.tools.backup.credentials import (
    CredentialAesKey,
    CredentialByteArray,
)
import atbu.tools.backup.credentials as credentials
from atbu.tools.backup.config_migration_helpers import (
    Version_001,
    set_password_to_keyring_001,
)
import atbu.tools.backup.config_migration_helpers as config_migration_helpers
from atbu.tools.backup.storage_def_credentials import StorageDefCredentialSet

Version001_storage_def_name1 = "my-cloud-backup-53fcc53f-a5a5-4939-a88a-e8ad1c95fb0e"

Version001_config = {
    "name": "ATBU Configuration",
    "version": "0.01",
    "general": {"backup-info-dir": "C:\\Users\\TestUser\\.atbu\\atbu-backup-info"},
    "storage-definitions": {
        Version001_storage_def_name1: {
            "interface": "someinterface",
            "provider": "some_storage_provider",
            "container": "my-cloud-backup-xyz",
            "driver": {
                "key": "some-storage-key",
                "project": "some-storage-project",
                "secret": "keyring",
            },
            "keyring-mapping": {
                "driver-secret": {
                    "service": Version001_storage_def_name1,
                    "username": "ATBU-storage-password",
                },
                "encryption-key": {
                    "service": Version001_storage_def_name1,
                    "username": "ATBU-backup-enc-key",
                },
            },
            "encryption": {"key": "keyring"},
        }
    },
}


@pytest.mark.parametrize(
    "is_backup_encryption,is_storage_secret,is_password_protected",
    [
        pytest.param(
            True,
            True,
            False,
            id="is_backup_encryption=True,is_storage_secret=True,is_password_protected=False",
        ),
        pytest.param(
            True,
            True,
            True,
            id="is_backup_encryption=True,is_storage_secret=True,is_password_protected=True",
        ),
        pytest.param(
            True,
            False,
            True,
            id="is_backup_encryption=True,is_storage_secret=False,is_password_protected=True",
        ),
        pytest.param(
            False,
            True,
            False,
            id="is_backup_encryption=False,is_storage_secret=True,is_password_protected=False",
        ),
    ],
)
def test_migrate_001_to_003(
    tmp_path: Path,
    pytester: Pytester,
    monkeypatch: MonkeyPatch,
    is_backup_encryption: bool,
    is_storage_secret: bool,
    is_password_protected: bool,
):
    if is_password_protected:

        cred_password_str = "test-cred-password$"
        cred_password = CredentialByteArray(cred_password_str.encode("utf-8"))

        def mock_prompt_password_return(
            prompt,
            prompt_again=None,
            hidden: bool = True,
        ):
            return CredentialByteArray(cred_password)

        monkeypatch.setattr(
            credentials,
            "prompt_for_password_with_yubikey_opt",
            mock_prompt_password_return,
        )

    test_atbu_cfg_001 = copy.deepcopy(Version001_config)
    storage_defs_section = test_atbu_cfg_001[
        Version_001.CONFIG_SECTION_STORAGE_DEFINITIONS
    ]
    storage_def_section = storage_defs_section[Version001_storage_def_name1]
    storage_def_mapping_section = storage_def_section[
        Version_001.CONFIG_SECTION_KEYRING_MAPPING
    ]
    storage_def_driver_section = storage_def_section[Version_001.CONFIG_SECTION_DRIVER]
    if not is_backup_encryption:
        del storage_def_section[Version_001.CONFIG_SECTION_ENCRYPTION]
        del storage_def_mapping_section[
            Version_001.CONFIG_SECTION_KEYRING_MAPPING_ENCRYPTION_KEY
        ]
    if not is_storage_secret:
        del storage_def_section[Version_001.CONFIG_SECTION_DRIVER]
        del storage_def_mapping_section[
            Version_001.CONFIG_SECTION_KEYRING_MAPPING_DRIVER_SECRET
        ]
    if len(storage_def_section[Version_001.CONFIG_SECTION_KEYRING_MAPPING]) == 0:
        del storage_def_section[Version_001.CONFIG_SECTION_KEYRING_MAPPING]

    AtbuConfig.get_user_default_config_dir().mkdir(parents=True, exist_ok=True)
    user_default_config_path = AtbuConfig.get_user_default_config_file_path()

    with open(user_default_config_path, "w", encoding="utf-8") as config_file:
        config_file.write(json.dumps(test_atbu_cfg_001, indent=4))

    cred_enc = None
    storage_secret = None
    cred_count = 0
    if is_backup_encryption:
        cred_count += 1
    if is_storage_secret:
        cred_count += 1

    if is_backup_encryption:
        cred_enc = CredentialAesKey()
        cred_enc.create_key()
        if is_password_protected:
            cred_enc.set(password=CredentialByteArray(cred_password))
            cred_enc.encrypt_key()
        set_password_to_keyring_001(
            service_name=Version001_storage_def_name1,
            username=Version_001.CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION,
            password_type=Version_001.CONFIG_PASSWORD_TYPE_ACTUAL,
            password_bytes=cred_enc.get_material_as_bytes(),
            password_is_base64=False,
        )
        if is_password_protected:
            cred_enc.decrypt_key()  # prepare for validation below.

    if is_storage_secret:
        storage_secret = CredentialByteArray("storage-secret-abc123$".encode("utf-8"))
        set_password_to_keyring_001(
            service_name=Version001_storage_def_name1,
            username=Version_001.CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD,
            password_type=Version_001.CONFIG_PASSWORD_TYPE_ACTUAL,
            password_bytes=storage_secret,
            password_is_base64=False,
        )

    storage_def_dict = copy.deepcopy(
        test_atbu_cfg_001[Version_001.CONFIG_SECTION_STORAGE_DEFINITIONS][
            Version001_storage_def_name1
        ]
    )
    cred_set = StorageDefCredentialSet(
        storage_def_name=Version001_storage_def_name1,
        storage_def_dict=storage_def_dict,
    )
    try:
        cred_set.populate()
        if is_storage_secret:
            pytest.fail(f"Populate of old 001 credentials should fail.")
        assert len(cred_set.storage_def_creds) == cred_count
    except BaseException:
        if not is_storage_secret:
            # Not a use case, but this state indicates mismatch between test assumption mismatch and
            # actual behavior.
            pytest.fail(
                f"Populate of old 001 credentials without storage secret should *not* fail."
            )
        assert len(cred_set.storage_def_creds) == 0

    # Re-get the overwritten config to trigger the migration.
    AtbuConfig.always_migrate = True

    # Perform the same as the old way: atbu_cfg = AtbuConfig.access_default_config()
    atbu_cfg = AtbuConfig(path=AtbuConfig.get_user_default_config_file_path())
    # pylint: disable=protected-access
    atbu_cfg._check_upgrade_default_config()

    atbu_cfg2: AtbuConfig
    atbu_cfg2, storage_def_name2, storage_def_dict2 = AtbuConfig.access_cloud_storage_config(
        storage_def_name=Version001_storage_def_name1,
        must_exist=True,
    )

    assert atbu_cfg2 is not None
    assert storage_def_name2 is not None
    assert storage_def_dict2 is not None
    assert atbu_cfg2.version == ATBU_CONFIG_FILE_VERSION_STRING_CURRENT
    assert storage_def_name2 == Version001_storage_def_name1
    assert isinstance(storage_def_dict2[CONFIG_VALUE_NAME_STORAGE_DEF_UNIQUE_ID], str)
    assert (len(storage_def_dict2[CONFIG_VALUE_NAME_STORAGE_DEF_UNIQUE_ID]) == len(str(uuid4())))
    assert (
        storage_def_section[CONFIG_VALUE_NAME_INTERFACE_TYPE]
        == storage_def_dict2[CONFIG_VALUE_NAME_INTERFACE_TYPE]
    )
    assert (
        storage_def_section[CONFIG_VALUE_NAME_PROVIDER]
        == storage_def_dict2[CONFIG_VALUE_NAME_PROVIDER]
    )
    assert (
        storage_def_section[CONFIG_VALUE_NAME_CONTAINER]
        == storage_def_dict2[CONFIG_VALUE_NAME_CONTAINER]
    )
    driver_section = storage_def_section.get(CONFIG_SECTION_DRIVER)
    driver_section2 = storage_def_dict2.get(CONFIG_SECTION_DRIVER)
    if is_storage_secret:
        assert isinstance(driver_section, dict)
        assert isinstance(driver_section2, dict)
        assert (
            driver_section[CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY]
            == driver_section2[CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY]
        )
        assert (
            driver_section[CONFIG_VALUE_NAME_DRIVER_STORAGE_PROJECT]
            == driver_section2[CONFIG_VALUE_NAME_DRIVER_STORAGE_PROJECT]
        )
        assert (
            driver_section[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET]
            == driver_section2[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET]
        )
        assert driver_section2[CONFIG_PASSWORD_TYPE] == CONFIG_PASSWORD_KIND_ACTUAL
    else:
        assert driver_section is None
        assert driver_section2 is None

    cred_set = StorageDefCredentialSet(
        storage_def_name=Version001_storage_def_name1,
        storage_def_dict=atbu_cfg2.get_storage_def_dict(
            storage_def_name=Version001_storage_def_name1,
            must_exist=True,
        ),
    )

    cred_set.populate()
    cred_set.unprotect()
    assert len(cred_set.storage_def_creds) == cred_count

    mig_desc_cred_encryption = cred_set.get_encryption_desc_cred()
    if is_backup_encryption:
        mig_cred_encryption = mig_desc_cred_encryption.credential
        assert mig_cred_encryption.the_key == cred_enc.the_key
    else:
        assert mig_desc_cred_encryption is None

    mig_desc_cred_storage = cred_set.get_storage_desc_cred()
    if is_storage_secret:
        mig_cred_storage = mig_desc_cred_storage.credential
        assert mig_cred_storage.the_key == storage_secret
    else:
        assert mig_desc_cred_storage is None



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=unused-variable
# pylint: disable=unused-import
# pylint: disable=wrong-import-position

import os
from pathlib import Path
import logging
from pytest import (
    LogCaptureFixture,
    CaptureFixture,
    fail,
    raises,
    Pytester,
    FixtureRequest,
    Config,
    RunResult,
    ExitCode,
)
import pytest

from atbu.tools.backup.config import AtbuConfig

from .common_helpers import (
    create_test_data_directory_minimal,
    create_test_data_directory_minimal_vary,
    establish_random_seed,
    create_test_data_directory_basic,
    DirInfo,
    extract_dir_info_from_verify_log,
    validate_backup_recovery,
    validate_backup_restore,
    validate_backup_restore_history,
    validate_cred_export_import,
    run_atbu,
    directories_match_entirely_by_path,
)

LOGGER = logging.getLogger(__name__)

ATBU_TEST_BACKUP_NAME = "AtbuTestBackup-5b497bb3-c9ef-48a9-af7b-2327fc17fb65"

# import pdb; pdb.set_trace()
# import pdb; pdb.set_trace()


def setup_module(module):  # pylint: disable=unused-argument
    pass


def teardown_module(module):  # pylint: disable=unused-argument
    pass


backup_restore_parameters = [
    pytest.param(
        "none",
        id="no_compression",
    ),
    pytest.param(
        "normal",
        id="normal_compression",
    ),
]


@pytest.mark.parametrize(
    "compression_type",
    backup_restore_parameters,
)
def test_backup_restore(
    compression_type,
    tmp_path: Path,
    pytester: Pytester,
):
    establish_random_seed(tmp_path)  # bytes([0,1,2,3])

    source_directory = tmp_path / "SourceDataDir"
    backup_directory = tmp_path / "BackupDestination"

    _, files_created = create_test_data_directory_basic(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)

    stdin_bytes = (
        f"{ATBU_TEST_BACKUP_NAME}{os.linesep}{os.linesep}{os.linesep}".encode()
    )

    validate_backup_restore(
        pytester=pytester,
        tmp_path=tmp_path,
        source_directory=source_directory,
        expected_total_files=total_files,
        storage_specifier=backup_directory,
        compression_type=compression_type,
        backup_timeout=60,
        restore_timeout=60,
        initial_backup_stdin=stdin_bytes,
    )
    pass  # pylint: disable=unnecessary-pass


def test_verify_digest_only(
    pytestconfig: Config,  # pylint: disable=unused-argument
    tmp_path: Path,
    capsys: CaptureFixture,  # pylint: disable=unused-argument
    caplog: LogCaptureFixture,  # pylint: disable=unused-argument
    pytester: Pytester,
):
    establish_random_seed(tmp_path)  # bytes([0,1,2,3])

    source_directory = tmp_path / "SourceDataDir"
    backup_directory = tmp_path / "BackupDestination"

    _, files_created = create_test_data_directory_basic(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)

    with DirInfo(source_directory) as source_dir_info:
        source_dir_info.gather_info(start_gathering_digests=True)
        assert len(source_dir_info.file_list) == total_files

        stdin_bytes = (
            f"{ATBU_TEST_BACKUP_NAME}{os.linesep}{os.linesep}{os.linesep}".encode()
        )
        rr = run_atbu(
            pytester,
            tmp_path,
            "backup",
            "--incremental-plus",
            source_directory,
            backup_directory,
            stdin=stdin_bytes,
            log_base_name="backup",
        )
        assert rr.ret == ExitCode.OK

        rr = run_atbu(
            pytester,
            tmp_path,
            "verify",
            backup_directory,
            "backup:last",
            "files:*",
            log_base_name="verify",
        )
        assert rr.ret == ExitCode.OK
        verify_dir_info = extract_dir_info_from_verify_log(output_lines=rr.outlines)
        assert directories_match_entirely_by_path(
            di1=source_dir_info, di2=verify_dir_info
        )
    pass  # pylint: disable=unnecessary-pass


def test_verify_compare_local(
    pytestconfig: Config,  # pylint: disable=unused-argument
    tmp_path: Path,
    capsys: CaptureFixture,  # pylint: disable=unused-argument
    caplog: LogCaptureFixture,  # pylint: disable=unused-argument
    pytester: Pytester,
):
    establish_random_seed(tmp_path)  # bytes([0,1,2,3])

    source_directory = tmp_path / "SourceDataDir"
    backup_directory = tmp_path / "BackupDestination"

    _, files_created = create_test_data_directory_basic(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)

    source_dir_info = DirInfo(source_directory)
    with source_dir_info:
        source_dir_info.gather_info(start_gathering_digests=True)
        assert len(source_dir_info.file_list) == total_files

        stdin_bytes = (
            f"{ATBU_TEST_BACKUP_NAME}{os.linesep}{os.linesep}{os.linesep}".encode()
        )
        rr = run_atbu(
            pytester,
            tmp_path,
            "backup",
            "--incremental-plus",
            source_directory,
            backup_directory,
            stdin=stdin_bytes,
            log_base_name="backup",
        )
        assert rr.ret == ExitCode.OK

        rr = run_atbu(
            pytester,
            tmp_path,
            "verify",
            backup_directory,
            "backup:last",
            "files:*",
            "--compare",
            log_base_name="verify-compare",
        )
        assert rr.ret == ExitCode.OK
        verify_dir_info = extract_dir_info_from_verify_log(output_lines=rr.outlines)
        assert directories_match_entirely_by_path(
            di1=source_dir_info, di2=verify_dir_info
        )
    pass  # pylint: disable=unnecessary-pass


@pytest.mark.parametrize(
    "compression_type",
    backup_restore_parameters,
)
def test_backup_restore_history(
    compression_type,
    tmp_path: Path,
    pytester: Pytester,
):
    establish_random_seed(tmp_path)  # bytes([0,1,2,3])

    source_directory = tmp_path / "SourceDataDir"
    backup_directory = tmp_path / "BackupDestination"

    _, files_created = create_test_data_directory_minimal_vary(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)

    stdin_bytes = (
        f"{ATBU_TEST_BACKUP_NAME}{os.linesep}{os.linesep}{os.linesep}".encode()
    )

    validate_backup_restore_history(
        pytester=pytester,
        tmp_path=tmp_path,
        max_history=5,
        source_directory=source_directory,
        expected_total_files=total_files,
        storage_specifier=backup_directory,
        compression_type=compression_type,
        backup_timeout=60,
        restore_timeout=60,
        initial_backup_stdin=stdin_bytes,
    )
    pass  # pylint: disable=unnecessary-pass


def test_credential_export_import(
    tmp_path: Path,
    pytester: Pytester,
):
    establish_random_seed(tmp_path)  # bytes([0,1,2,3])

    storage_def_name = ATBU_TEST_BACKUP_NAME

    source_directory = tmp_path / "SourceDataDir"
    backup_directory = tmp_path / "BackupDestination"

    _, files_created = create_test_data_directory_minimal(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)
    assert total_files > 0

    stdin_bytes = f"{storage_def_name}{os.linesep}{os.linesep}{os.linesep}".encode()
    rr = run_atbu(
        pytester,
        tmp_path,
        "backup",
        "--full",
        source_directory,
        backup_directory,
        stdin=stdin_bytes,
        log_base_name="backup",
    )
    assert rr.ret == ExitCode.OK

    atbu_cfg: AtbuConfig
    (
        atbu_cfg,
        storage_def_name_from_cfg,
        storage_def,
    ) = AtbuConfig.access_filesystem_storage_config(
        storage_location_path=backup_directory,
        resolve_storage_def_secrets=False,
        create_if_not_exist=False,
        prompt_to_create=False,
    )
    assert storage_def_name.lower() == storage_def_name_from_cfg

    validate_cred_export_import(
        pytester=pytester,
        tmp_path=tmp_path,
        atbu_cfg_path=atbu_cfg.path,
        source_directory=source_directory,
        expected_total_files=total_files,
        storage_def_name=storage_def_name,
        storage_specifier=backup_directory,
    )

    # Delete the secrets from potentially global cred store, but leave
    # the def intact in case it needs to be examined.
    # atbu_cfg.delete_storage_def(storage_def_name=storage_def_name)
    atbu_cfg.delete_storage_def_secrets(storage_def_name=storage_def_name)
    pass  # pylint: disable=unnecessary-pass


def test_credential_and_backup_info_recovery(
    tmp_path: Path,
    pytester: Pytester,
):
    establish_random_seed(tmp_path)  # bytes([0,1,2,3])

    storage_def_name = ATBU_TEST_BACKUP_NAME

    source_directory = tmp_path / "SourceDataDir"
    backup_directory = tmp_path / "BackupDestination"

    _, files_created = create_test_data_directory_minimal(
        path_to_dir=source_directory,
    )
    total_files = len(files_created)
    assert total_files > 0

    stdin_bytes = f"{storage_def_name}{os.linesep}{os.linesep}{os.linesep}".encode()
    rr = run_atbu(
        pytester,
        tmp_path,
        "backup",
        "--full",
        source_directory,
        backup_directory,
        stdin=stdin_bytes,
        log_base_name="backup",
    )
    assert rr.ret == ExitCode.OK

    atbu_cfg: AtbuConfig
    (
        atbu_cfg,
        storage_def_name_from_cfg,
        storage_def,
    ) = AtbuConfig.access_filesystem_storage_config(
        storage_location_path=backup_directory,
        resolve_storage_def_secrets=False,
        create_if_not_exist=False,
        prompt_to_create=False,
    )
    assert storage_def_name.lower() == storage_def_name_from_cfg

    validate_backup_recovery(
        pytester=pytester,
        tmp_path=tmp_path,
        atbu_cfg_path=atbu_cfg.path,
        source_directory=source_directory,
        expected_total_files=total_files,
        storage_def_name=storage_def_name,
        storage_specifier=backup_directory,
    )

    # Delete the secrets from potentially global cred store, but leave
    # the def intact in case it needs to be examined.
    # atbu_cfg.delete_storage_def(storage_def_name=storage_def_name)
    atbu_cfg.delete_storage_def_secrets(storage_def_name=storage_def_name)
    pass  # pylint: disable=unnecessary-pass



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Persistent file information database-related functionality allows
gathering of persistent file information, placing it into a single
json "database" file.

There was a second revision of persistent info functionality incorporated
into ATBU so you may see redundancy or superfluous constructs within.
"""

from dataclasses import dataclass
import os
import json
import logging
from collections import defaultdict

from atbu.common.hasher import HasherDefinitions

from ..backup.global_hasher import GlobalHasherDefinitions
from ..backup.exception import *
from ..backup.constants import *
from .file_info import (
    FileInformationPersistent,
    get_loc_persistent_file_info,
    LocationFileInfoUpdater,
)


class CustomEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, FileInformationPersistent):
            return o.to_serialization_dict()
        return json.JSONEncoder.default(self, o)


class CustomDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):  # pylint: disable=method-hidden,no-self-use
        if "_type" not in obj:
            return obj
        object_type = obj["_type"]
        if object_type == "FileInformationPersistent":
            return FileInformationPersistent(
                path=obj["path"],
                info_current=obj["info_current"],
                info_history=obj["info_history"],
                is_loaded_from_db=True,
            )
        elif (
            object_type == "FileInformation"
        ):  # FileInformationPersistent used to be FileInformation (when used w/.atbu files)
            return FileInformationPersistent(
                path=obj["path"],
                info_current=obj["info_current"],
                info_history=obj["info_history"],
                is_loaded_from_db=True,
            )
        return obj


def rel_path(top_level_dir: str, path: str):
    check_top = os.path.normcase(path[: len(top_level_dir)])
    if os.path.normcase(top_level_dir) != check_top:
        raise InvalidFunctionArgument(
            f"The path must be a subdirectory of top_level_dir: "
            f"top_level_dir={top_level_dir} path={path}"
        )
    rpath = path[len(top_level_dir) :]
    if rpath[0] in [os.sep, os.altsep]:
        rpath = rpath[1:]
    return rpath


def flatten_location_info_to_path_sorted_list(
    location_info: dict[str, list[FileInformationPersistent]]
) -> list[FileInformationPersistent]:
    return sorted(
        [file_info for fi_list in location_info.values() for file_info in fi_list],
        key=lambda x: x.nc_path,
    )


def get_location_info_from_database(
    database_json_path,
) -> defaultdict[str, list[FileInformationPersistent]]:
    with open(database_json_path, "r", encoding="utf-8") as database_file:
        return json.load(database_file, cls=CustomDecoder)


def file_info_list_to_digest_dict(
    file_info_list: list[FileInformationPersistent],
):
    digest_dict: defaultdict[str, list[FileInformationPersistent]] = defaultdict(
        list[FileInformationPersistent]
    )
    for file_info in file_info_list:
        digest_dict[file_info.get_current_digest()].append(file_info)
    return digest_dict


def file_info_list_to_ncpath_dict(
    file_info_list: list[FileInformationPersistent],
) -> dict[FileInformationPersistent]:
    ncpath_dict: defaultdict[str, FileInformationPersistent] = {}
    for file_info in file_info_list:
        fi = ncpath_dict.get(file_info.nc_path)
        if fi:
            raise InvalidStateError(
                f"Expecting only one file info per path: path={fi.path}"
            )
        ncpath_dict[file_info.nc_path] = file_info
    return ncpath_dict


def save_database_from_dict_file_info(
    dict_digest_to_file_info_list: defaultdict[str, list[FileInformationPersistent]],
    database_path,
):
    with open(database_path, "w", encoding="utf-8") as database_file:
        database_file.write(
            json.dumps(dict_digest_to_file_info_list, cls=CustomEncoder, indent=4)
        )


def save_database_from_file_info_list(
    file_info_list: list[FileInformationPersistent],
    database_path,
    hasher_defs: HasherDefinitions = None,
):
    if not hasher_defs:
        hasher_defs = GlobalHasherDefinitions()
    digest_dict = file_info_list_to_digest_dict(
        file_info_list=file_info_list,
    )
    save_database_from_dict_file_info(
        dict_digest_to_file_info_list=digest_dict, database_path=database_path
    )


class FileInformationDatabase:
    def __init__(self, source_path: str, persist_type: ATBU_PERSIST_TYPE_HINT) -> None:
        if not os.path.exists(source_path):
            raise InvalidFunctionArgument(
                f"Source path must a be directory but got '{source_path}'."
            )
        self.source_path = os.path.abspath(source_path)
        # If is_source_path_explicit_json_db_file==True means .json is user-created
        # and could pertain to one or more directories not currently online.
        self.is_source_path_explicit_json_db_file = os.path.isfile(self.source_path)
        self.persist_type = persist_type
        if self.is_source_path_explicit_json_db_file:
            self.db_filename = self.source_path
        else:
            self.db_filename = os.path.join(
                self.source_path, ATBU_DEFAULT_PERSISTENT_DB_FILENAME
            )
        self._file_info_list: list[FileInformationPersistent] = None
        self.updater: LocationFileInfoUpdater = None

    def get_file_info(self) -> list[FileInformationPersistent]:
        if self._file_info_list is None:
            raise InvalidStateError(
                f"_file_info_list is None. You must first either load or update the DB."
            )
        return self._file_info_list

    def get_dict_digest_to_fi(self):
        return file_info_list_to_digest_dict(file_info_list=self.get_file_info())

    def get_dict_ncpath_to_fi(self):
        return file_info_list_to_ncpath_dict(file_info_list=self.get_file_info())

    def _get_source_path_file_info_list(self):
        return get_loc_persistent_file_info(locations=self.source_path)

    def _load_per_file_info(self) -> None:
        if self.persist_type != ATBU_PERSIST_TYPE_PER_FILE:
            raise InvalidFunctionArgument(
                f"Expected persist_type={ATBU_PERSIST_TYPE_PER_FILE} but got {self.persist_type}"
            )
        self._file_info_list = self._get_source_path_file_info_list()

    def _load_per_dir_info(self) -> None:
        """Load the .json DB file. Two types of .json DB files based on
        how they were created/used. An user-created DB is created with
        save-db and is not tied to a specific directory, can contain
        files from one or more directories (up to user). The other
        type of DB is that created by update-digests for per-dir cases,
        where it maintains a DB of files within that directory structure.
        This method will load either type and the result is a flat list
        of file information in self._file_info_list.
        """
        if self.persist_type != ATBU_PERSIST_TYPE_PER_DIR:
            raise InvalidStateError(
                f"Expected persist_type={ATBU_PERSIST_TYPE_PER_DIR} but got {self.persist_type}"
            )

        if self.db_filename is None:
            raise InvalidStateError(
                f"db_filename is None, expected db_filename to be a .json db filename."
            )

        # Load file information from the DB file.
        # The info from DB contains path, digests,
        # modified date/time, size, etc.
        digest_to_fi_list = {}
        if os.path.exists(self.db_filename):
            digest_to_fi_list = get_location_info_from_database(
                database_json_path=self.db_filename
            )
        # Flatten to list of file info.
        self._file_info_list = [fi for l in digest_to_fi_list.values() for fi in l]

        # Disallow refresh of file information that comes from
        # an explicitly specified .json DB. This will cause
        # an exception if there's a code logic breaking the
        # assumption we do not expect files to be present
        # with an explicitly specified database.
        if self.is_source_path_explicit_json_db_file:
            for fi in self._file_info_list:
                fi.implicit_refresh_allowed = False

        # If user specifies explicit .json db as a location,
        # the user expects the db to be used as is with the need
        # for files to be present (user captured db earlier, perhaps,
        # before a drive was disconnected). Therefore, skip the
        # following which assumes files exist on the system.
        if not self.is_source_path_explicit_json_db_file:
            # Remove file info for files that no longer exist.
            # Or, in the case of a moved folder, remove info
            # pertaining to the old location.
            nc_source_path = os.path.normcase(self.source_path)
            updated_file_info_list = []
            for fi in self._file_info_list:
                is_root_valid = False
                try:
                    common_prefix = os.path.commonpath([nc_source_path, fi.nc_path])
                    is_root_valid = common_prefix == nc_source_path
                except ValueError:
                    pass
                if is_root_valid and os.path.isfile(fi.path):
                    updated_file_info_list.append(fi)
            self._file_info_list = updated_file_info_list

            #
            # self._file_info_list is now a file info list of files
            # that both exist and are under self.source_path.
            #

            # Index self._file_info_list by path which will be used
            # to find out what file information is missing.
            db_path_to_file_info: dict = {fi.nc_path: fi for fi in self._file_info_list}

            # Get current file system file info. This only gets
            # a bare file info without digest/stat (because they
            # are not updated yet as DB is before being saved).
            fs_file_info_list = self._get_source_path_file_info_list()

            # Index new paths into the self._file_info_list DB info.
            for fi in fs_file_info_list:
                # If not in DB, add to DB.
                if not db_path_to_file_info.get(fi.nc_path):
                    db_path_to_file_info[fi.nc_path] = fi

            self._file_info_list = list(db_path_to_file_info.values())

    def load(self):
        """This loads existing data. TBD relationship with _get_source_path_file_info
        et al which scan files for both persist_types (in prep of updates etc.)
        """
        if self.persist_type == ATBU_PERSIST_TYPE_PER_FILE:
            self._load_per_file_info()
        elif self.persist_type == ATBU_PERSIST_TYPE_PER_DIR:
            self._load_per_dir_info()
        else:
            raise InvalidFunctionArgument(
                f"Unexpected persist_type={self.persist_type}"
            )

    def update(
        self,
        change_detection_type: str,
        update_stale: bool,
        whatif: bool,
    ) -> LocationFileInfoUpdater:
        self.load()
        # Show progress in sorted file path order.
        self._file_info_list.sort(key=lambda fi: fi.nc_path)
        self.updater = LocationFileInfoUpdater(
            file_info_list=self._file_info_list,
            update_stale=update_stale,
            change_detection_type=change_detection_type,
            per_file_persistence=not self.persist_type == ATBU_PERSIST_TYPE_PER_DIR,
            whatif=whatif,
        )
        # Do not update for explicitly specified .json
        # file DBs because files are not necessarily online.
        if not self.is_source_path_explicit_json_db_file:
            self.updater.update()
        return self.updater

    def update_from(self, other_db):
        self._file_info_list = list(other_db.get_file_info())
        # If this instance is a 'per-dir' DB, save the DB.
        # There is no saving to perform for 'per-file' since
        # per-file persistence updates as part of scanning/updating
        # the directories, likely before this update_from is called.
        if self.persist_type == ATBU_PERSIST_TYPE_PER_DIR:
            self.save_db()

    def save_db(self):
        save_database_from_file_info_list(
            database_path=self.db_filename,
            file_info_list=self._file_info_list,
        )


class FileInformationDatabaseCollection:
    def __init__(self, source_path: str, persist_types: list[str]) -> None:
        if not os.path.exists(source_path):
            raise InvalidFunctionArgument(
                f"Source path must a be directory but got '{source_path}'."
            )
        self.source_path = os.path.abspath(source_path)
        self.databases = {}
        for pt in persist_types:
            if pt not in ATBU_PERSIST_TYPES:
                raise InvalidFunctionArgument(
                    f"Expected a valid persist type '{ATBU_PERSIST_TYPES}' but got '{pt}'."
                )
            if self.databases.get(pt):
                raise InvalidFunctionArgument(
                    f"Only one of each persist type is allowed but got more than one of '{pt}'."
                )
            self.databases[pt] = FileInformationDatabase(
                persist_type=pt, source_path=self.source_path
            )

    @property
    def has_per_file_persistence(self):
        for pt in self.databases:
            if pt == ATBU_PERSIST_TYPE_PER_FILE:
                return True
        return False

    def get_combined_db_file_info(self):
        combined_info: dict[str, FileInformationPersistent] = {}
        db: FileInformationDatabase
        for k in [ATBU_PERSIST_TYPE_PER_FILE, ATBU_PERSIST_TYPE_PER_DIR]:
            db = self.databases.get(k)
            if not db:
                continue
            fi_list = db.get_file_info()
            for fi in fi_list:
                if not combined_info.get(fi.nc_path):
                    combined_info[fi.nc_path] = fi
        return combined_info.values()

    def get_dict_nc_path_to_fi(self):
        # pylint: disable=unused-variable
        (
            primary_update_db,
            other_db,
        ) = self._get_primary_other_db()  # pylint: disable=unused-variable
        return primary_update_db.get_dict_ncpath_to_fi()

    def get_dict_digest_to_fi(self):
        # pylint: disable=unused-variable
        (
            primary_update_db,
            other_db,
        ) = self._get_primary_other_db()  # pylint: disable=unused-variable
        return primary_update_db.get_dict_digest_to_fi()

    def _get_primary_other_db(self):
        primary_update_db: FileInformationDatabase = self.databases.get(
            ATBU_PERSIST_TYPE_PER_FILE
        )
        other_db: FileInformationDatabase = self.databases.get(
            ATBU_PERSIST_TYPE_PER_DIR
        )
        if not primary_update_db:
            primary_update_db = self.databases.get(ATBU_PERSIST_TYPE_PER_DIR)
            other_db = None
        if primary_update_db is None:
            raise InvalidFunctionArgument(
                f"Cannot determine the primary update db. DBs={self.databases.keys()}"
            )
        return primary_update_db, other_db

    def load(self):
        primary_update_db, other_db = self._get_primary_other_db()
        if primary_update_db is not None:
            primary_update_db.load()
        if other_db is not None:
            other_db.load()

    def update(
        self,
        change_detection_type: str,
        update_stale: bool,
        whatif: bool,
    ) -> LocationFileInfoUpdater:
        """Update the databases if needed. Returns the updater used for
        access to any stats.
        """
        primary_update_db, other_db = self._get_primary_other_db()
        updater = primary_update_db.update(
            change_detection_type=change_detection_type,
            update_stale=update_stale,
            whatif=whatif,
        )
        if other_db is None:
            # There is no other db, just primary. If primary is per-file it
            # is implicitly already persisted given the update above. If it
            # is a per-dir, it must be saved after the update.
            # Save the per-dir db.
            if (
                primary_update_db.persist_type == ATBU_PERSIST_TYPE_PER_DIR
                and not primary_update_db.is_source_path_explicit_json_db_file
            ):
                primary_update_db.save_db()
        else:
            # Tell other_db to update itself from
            # primary_update_db and save itself to db.
            other_db.update_from(other_db=primary_update_db)
        return updater


def extract_location_info(
    arguments: list[str], min_required: int, max_allowed: int, must_exist: bool = True,
) -> list[tuple[str, str]]:
    """Helper function to process location arguments captured by argparse."""
    per_arg_to_persist_types = {
        "per-file:": [ATBU_PERSIST_TYPE_PER_FILE],
        "pf:": [ATBU_PERSIST_TYPE_PER_FILE],
        "per-dir:": [ATBU_PERSIST_TYPE_PER_DIR],
        "pd:": [ATBU_PERSIST_TYPE_PER_DIR],
        "per-both:": [ATBU_PERSIST_TYPE_PER_FILE, ATBU_PERSIST_TYPE_PER_DIR],
        "pb:": [ATBU_PERSIST_TYPE_PER_FILE, ATBU_PERSIST_TYPE_PER_DIR],
    }
    current_persist_types = [ATBU_PERSIST_TYPE_PER_DIR]
    locations = []
    for loc in arguments:
        orig_loc = loc
        if max_allowed is not None and len(locations) >= max_allowed:
            raise InvalidCommandLineArgument(
                f"You have already specified {len(locations)} locations. "
                f"The argument '{loc}' is invalid."
            )
        if loc in per_arg_to_persist_types:
            current_persist_types = per_arg_to_persist_types[loc]
            continue
        if len(loc) == 0:
            raise InvalidCommandLineArgument(
                f"Invalid argument '{orig_loc}'. Expected a directory but got '{loc}'"
            )
        loc = os.path.abspath(loc)
        if must_exist and not os.path.exists(loc):
            raise InvalidCommandLineArgument(
                f"The specified location does not exist: {loc}"
            )
        locations.append(
            (
                loc,
                current_persist_types,
            )
        )
    if len(locations) < min_required:
        raise InvalidCommandLineArgument(
            f"You have only specified '{len(locations)}' "
            f"which is less than the {min_required} required."
        )
    if max_allowed is not None and len(locations) > max_allowed:
        # Given above checks, this should never occur.
        raise InvalidCommandLineArgument(
            f"You specified {len(locations)} locations which "
            f"is more than the {max_allowed} allowed."
        )
    return locations


@dataclass
class LocationSummaryItem:
    location: str
    total_all_files: int
    total_found_unique_files: int
    total_found_physical_files: int
    total_skipped_files: int


def log_location_summary_item(lsi: LocationSummaryItem):
    logging.info(f"Location .............................. {lsi.location}")
    logging.info(f"Location total all files .............. {lsi.total_all_files}")
    logging.info(
        f"Location total found unique files ..... {lsi.total_found_unique_files}"
    )
    logging.info(
        f"Location total found physical files ... {lsi.total_found_physical_files}"
    )
    if lsi.total_skipped_files > 0:
        logging.warning(
            f"Location total skipped files ....... {lsi.total_skipped_files}"
        )
    else:
        logging.info(
            f"Location total skipped files .......... {lsi.total_skipped_files}"
        )


def handle_savedb(args):
    logging.debug("handle_savedb func")
    if args.locations is None:
        raise InvalidCommandLineArgument("You must specify one or more locations.")
    logging.debug(f"locations={args.locations}")
    location_persist_types = extract_location_info(
        args.locations, min_required=1, max_allowed=None
    )

    if not hasattr(args, "database"):
        raise InvalidCommandLineArgument(
            "The --database location has not been specified."
        )
    database_path = args.database
    database_skipped_path = os.path.splitext(database_path)[0] + "-skipped.txt"
    logging.info(f"Database: {database_path}")
    if os.path.exists(database_path):
        if not args.overwrite:
            raise InvalidCommandLineArgument(
                f"The database file already exists: {database_path}"
            )
        os.remove(database_path)
        logging.info(f"Database file erased: {database_path}")
    if os.path.exists(database_skipped_path):
        if not args.overwrite:
            raise InvalidCommandLineArgument(
                f"The database skipped list file already exists: {database_skipped_path}"
            )
        os.remove(database_skipped_path)
        logging.info(f"Database skipped file erased: {database_skipped_path}")
    all_locations_info: defaultdict[str, list[FileInformationPersistent]] = defaultdict(
        list[FileInformationPersistent]
    )
    all_locations_info_skipped = []
    location_summary = []
    for lp in location_persist_types:
        location = lp[0]
        persist_types = lp[1]
        if not os.path.exists(location):
            raise FileNotFoundError(f"Directory does not exist: {location}")

        location_DBs = FileInformationDatabaseCollection(
            source_path=location, persist_types=persist_types
        )
        updater = location_DBs.update(
            change_detection_type=args.change_detection_type,
            update_stale=args.update_stale,
            whatif=False,
        )
        location_info_found = location_DBs.get_dict_digest_to_fi()
        location_info_skipped = updater.skipped_files

        lsi = LocationSummaryItem(
            location=location,
            total_all_files=len(location_info_found) + len(location_info_skipped),
            total_found_unique_files=len(location_info_found),
            total_found_physical_files=sum(
                len(l) for l in location_info_found.values()
            ),
            total_skipped_files=len(location_info_skipped),
        )
        location_summary.append(lsi)
        log_location_summary_item(lsi=lsi)
        for k, v in location_info_found.items():
            all_locations_info[k] += v
        all_locations_info_skipped += location_info_skipped
    save_database_from_dict_file_info(
        dict_digest_to_file_info_list=all_locations_info, database_path=database_path
    )
    with open(database_skipped_path, "w", encoding="utf-8") as database_skipped_file:
        database_skipped_file.write(json.dumps(all_locations_info_skipped))
    logging.info(
        f"========================================================================="
    )
    logging.info(f"The following is a recap of summary information output above:")
    for lsi in location_summary:
        log_location_summary_item(lsi=lsi)
        logging.info(
            f"-------------------------------------------------------------------------"
        )
    logging.info(f"All locations total unique files ...... {len(all_locations_info)}")
    logging.info(
        f"All locations total physical files .... "
        f"{sum(len(l) for l in all_locations_info.values())}"
    )
    if len(all_locations_info_skipped) > 0:
        logging.warning(
            f"All locations skipped files ........... {len(all_locations_info_skipped)}"
        )
    else:
        logging.info(
            f"All locations skipped files ........... {len(all_locations_info_skipped)}"
        )



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Backup-related command line handlers.
"""

import os
from pathlib import Path
import logging
from typing import Union

from atbu.common.util_helpers import (
    convert_to_pathlib_path,
    is_absolute_path,
)

from .constants import *
from .exception import *

from .config import (
    AtbuConfig,
    is_storage_def_name_ok,
    parse_storage_def_specifier,
)
from .storage_def_credentials import StorageDefCredentialSet
from .creds_cmdline import setup_backup_encryption_wizard
from .backup_selections import get_local_file_information
from .backup_core import Backup, StorageDefinition


def get_local_filesystem_backup_info(
    storage_location_path: Union[str, Path],
    resolve_storage_def_secrets: bool = False,
    create_if_not_exist: bool = False,
    prompt_to_create: bool = False,
):
    storage_location_path = convert_to_pathlib_path(storage_location_path)
    if storage_location_path is None or storage_location_path.is_file():
        raise ValueError(
            "The backup storage location must be an absolute path and not an existing file."
        )
    if not is_absolute_path(path_to_dir=storage_location_path):
        raise ValueError("The backup storage location must be an absolute path.")

    storage_atbu_cfg: AtbuConfig
    (
        storage_atbu_cfg,
        storage_def_name,
        storage_def,
    ) = AtbuConfig.access_filesystem_storage_config(
        storage_location_path=storage_location_path,
        resolve_storage_def_secrets=resolve_storage_def_secrets,
        create_if_not_exist=create_if_not_exist,
        prompt_to_create=prompt_to_create,
    )
    return storage_atbu_cfg, storage_def_name, storage_def


def get_local_filesystem_backup_with_wizard(
    storage_location_path: Union[str, Path],
    default_storage_def_name: str = None,
) -> tuple[AtbuConfig, str, dict]:
    storage_location_path = convert_to_pathlib_path(storage_location_path)
    if storage_location_path is None or storage_location_path.is_file():
        raise ValueError(
            "The backup storage location must be an absolute path and not an existing file."
        )
    if not is_absolute_path(path_to_dir=storage_location_path):
        raise ValueError("The backup storage location must be an absolute path.")

    storage_atbu_cfg, storage_def_name, storage_def = get_local_filesystem_backup_info(
        storage_location_path=storage_location_path,
        create_if_not_exist=True,
        prompt_to_create=False,
    )
    print(f"Storage location: {str(storage_location_path)}")
    print(f"Storage definition: {str(storage_atbu_cfg.path)}")
    if storage_def is not None:
        return storage_atbu_cfg, storage_def_name, storage_def

    if not default_storage_def_name:
        default_storage_def_name = os.path.split(storage_location_path)[1]
    default_storage_def_name = default_storage_def_name.lower()
    print(
        f"Backup destinations require a storage definition which retains information about the"
    )
    print(
        f"storage location, including how to access it and whether it's cloud or filesystem-based."
    )
    print(
        f"Enter a user-friendly name for this backup destination's storage definition."
    )
    print(f"Any name you enter will be converted to all lower case.")
    print(
        f"If you press ENTER without entering anything, '{default_storage_def_name}' will be used."
    )
    while True:
        storage_def_name = input("Enter a name (letters, numbers, spaces): ")
        if len(storage_def_name) == 0:
            print(f"Using '{default_storage_def_name}'.")
            storage_def_name = default_storage_def_name
        storage_def_name = storage_def_name.lower()
        if AtbuConfig.is_user_storage_def_exists(storage_def_name=storage_def_name):
            print(f"That name already exists, try another name.")
            continue
        if not is_storage_def_name_ok(storage_def_name=storage_def_name):
            print(f"Invalid character(s).")
            continue
        print(f"Using the name '{storage_def_name}'...")
        break
    print(f"Creating backup storage definition...")
    storage_def_name, storage_def = storage_atbu_cfg.create_filesystem_storage_def(
        storage_location_path=storage_location_path,
        unique_storage_def_name=storage_def_name,
    )
    print(f"Created storage definition {storage_def_name} for {storage_location_path}")
    desc_cred = setup_backup_encryption_wizard(
        storage_atbu_cfg=storage_atbu_cfg,
        storage_def_name=storage_def_name,
    )
    cred_set = StorageDefCredentialSet(
        storage_def_name=storage_def_name,
        storage_def_dict=storage_def,
    )
    cred_set.append(
        desc_cred=desc_cred,
        affected_config_path_parts=CRED_SECRET_KIND_ENCRYPTION.split("-"),
    )
    print(f"Storing...")
    cred_set.protect()
    cred_set.save()
    print(f"Your key is stored.")
    print(f"Saving {storage_atbu_cfg.path}")
    storage_atbu_cfg.save_config_file()
    print(f"{storage_atbu_cfg.path} has been saved.")

    return storage_atbu_cfg, storage_def_name, storage_def


def establish_storage_path_location():
    while True:
        print(
            f"Enter an absolute path to directory to act as the storage location for the backup."
        )
        print(f"The directory should either be non-existent or empty.")
        storage_location_path = input(f"Enter storage location path:")
        if (
            os.path.exists(storage_location_path)
            and len(os.listdir(storage_location_path)) != 0
        ):
            print(
                f"The specified storage location path exists and is not empty, try again."
            )
            continue
        return storage_location_path


def handle_new_local_filesystem_storage_def(
    default_storage_def_name: str, storage_location: str
) -> tuple[AtbuConfig, str, dict]:

    if not storage_location:
        storage_location = establish_storage_path_location()
    #
    # Local filesystem storage:
    #
    (
        atbu_cfg_to_use,
        storage_def_name,
        storage_def,
    ) = get_local_filesystem_backup_with_wizard(
        storage_location_path=storage_location,
        default_storage_def_name=default_storage_def_name,
    )

    return atbu_cfg_to_use, storage_def_name, storage_def


def handle_backup(args):
    backup_type = None
    sneaky_corruption_detection: bool = args.detect_bitrot
    if args.full:
        backup_type = ATBU_BACKUP_TYPE_FULL
    elif args.incremental:
        backup_type = ATBU_BACKUP_TYPE_INCREMENTAL
    elif args.ip:
        backup_type = ATBU_BACKUP_TYPE_INCREMENTAL_PLUS
    else:
        raise ValueError(f"Could not derive backup type.")

    deduplication_option = None
    if args.dedup is not None:
        if backup_type != ATBU_BACKUP_TYPE_INCREMENTAL_PLUS:
            raise InvalidCommandLineArgument(
                f"The --dedup option is only valid with --incremental-plus"
            )
        deduplication_option = args.dedup[0]

    source_locations: list[str] = args.source_dirs
    dest_location: str = args.dest_storage_specifier
    exclude_patterns: list[str] = args.exclude

    logging.debug(f"Backup type: {backup_type}")
    logging.debug(f"Source locations: {source_locations}")
    logging.debug(f"Dest location: {dest_location}")
    logging.debug(f"Exclude patterns: {exclude_patterns}")
    logging.debug(f"sneaky_corruption_detection: {sneaky_corruption_detection}")

    #
    # The config to use is the one relating to the backup destination.
    # If it's a filesystem local disk then atbu_cfg_to_use is usually
    # the ATBU config file found in that disk's .atbu folder, otherwise
    # it's a storage config in the user's default config directory.
    #
    atbu_cfg_to_use: AtbuConfig = None

    storage_def_name = parse_storage_def_specifier(storage_location=dest_location)
    if storage_def_name:
        #
        # A "storage:<storage_def_name>" was specified.
        # The storage_def_name may or may not exist,
        # and if it exists, may be either cloud/local.
        #
        atbu_cfg_to_use, _, storage_def_dict = AtbuConfig.access_cloud_storage_config(
            storage_def_name=storage_def_name,
            must_exist=False,
            create_if_not_exist=False,
        )
        if not atbu_cfg_to_use or not storage_def_dict:
            # Does not exist yet.
            raise StorageDefinitionNotFoundError(
                f"The storage definition '{storage_def_name}' was not found. "
                f"You can create a storage definition using '{ATBU_PROGRAM_NAME} "
                f"creds {CREDS_SUBCMD_CREATE_STORAGE_DEF}...'."
            )
    else:
        #
        # Local filesystem storage:
        #
        atbu_cfg_to_use, storage_def_name, _ = handle_new_local_filesystem_storage_def(
            default_storage_def_name=None, storage_location=dest_location
        )

    backup_info_dirs: list[str] = [str(bid) for bid in atbu_cfg_to_use.get_backup_info_dirs()]

    storage_def_dict = atbu_cfg_to_use.get_storage_def_with_resolved_secrets_deep_copy(
        storage_def_name=storage_def_name
    )

    storage_def = StorageDefinition.storage_def_from_dict(
        storage_def_name=storage_def_name, storage_def_dict=storage_def_dict
    )

    logging.info(f"Backup location(s)...")
    for idx, source_location in enumerate(source_locations):
        idx_str = f"  Source location #{idx} "
        logging.info(f"{idx_str:.<35} {source_location}")

    logging.info(f"Searching for files...")
    file_info_list = get_local_file_information(
        src_dirs_wc=source_locations, exclude_patterns=exclude_patterns
    )
    if len(file_info_list) == 0:
        logging.info(f"No files found, nothing to backup.")
        return

    compression_settings = atbu_cfg_to_use.get_compression_settings_deep_copy(
        storage_def_name=storage_def_name,
    )
    if args.compression is not None:
        compression_settings[CONFIG_VALUE_NAME_COMPRESSION_LEVEL] = args.compression

    logging.info(f"Backup destination: {dest_location}")
    with Backup(
        backup_type=backup_type,
        deduplication_option=deduplication_option,
        compression_settings=compression_settings,
        sneaky_corruption_detection=sneaky_corruption_detection,
        primary_backup_info_dir=backup_info_dirs[0],
        secondary_backup_info_dirs=backup_info_dirs[1:],
        source_file_info_list=file_info_list,
        storage_def=storage_def,
    ) as backup:
        backup.backup_files()

    if backup.is_completely_successful():
        logging.info(f"Success, no errors detected.")
        return 0
    else:
        logging.info(
            f"Some errors were detected. See prior messages and/or logs for details."
        )
        return 1
